import pytest

from icupy.icu import U_ICU_VERSION_MAJOR_NUM

if U_ICU_VERSION_MAJOR_NUM < 73:
    pytest.skip("ICU4C<73", allow_module_level=True)

# fmt: off
from icupy.icu import (
    DecimalFormatSymbols, ICUError, Locale, UErrorCode,
    UNumberFormatRoundingMode, UNumberGroupingStrategy, USimpleNumberSign,
)
from icupy.icu.number import (
    FormattedNumber, SimpleNumber, SimpleNumberFormatter,
)

# fmt: on


def test_api():
    # icu::number::SimpleNumberFormatter::SimpleNumberFormatter()
    fmt1 = SimpleNumberFormatter()

    # FormattedNumber icu::number::SimpleNumberFormatter::formatInt64(
    #       int64_t value,
    #       UErrorCode &status
    # ) const
    with pytest.raises(ICUError) as exc_info:
        _ = fmt1.format_int64(1)  # Default constructed formatter
    assert exc_info.value.args[0] == UErrorCode.U_INVALID_STATE_ERROR


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM > 74, reason="ICU4C>74")
def test_deprecated_75():
    """Deprecated or Obsoleted in ICU 75."""
    snf = SimpleNumberFormatter.for_locale_and_grouping_strategy(
        "en-US", UNumberGroupingStrategy.UNUM_GROUPING_ON_ALIGNED
    )

    # void icu::number::SimpleNumber::truncateStart(
    #       uint32_t maximumIntegerDigits,
    #       UErrorCode &status
    # )
    num = SimpleNumber.for_int64(918273645)
    num.multiply_by_power_of_ten(-4)
    num.truncate_start(4)

    result = snf.format(num)
    assert result.to_temp_string() == "1,827.3645"

    with pytest.raises(ICUError) as exc_info:
        _ = snf.format(num)  # Use of moved number
    assert exc_info.value.args[0] == UErrorCode.U_ILLEGAL_ARGUMENT_ERROR


def test_for_locale():
    # static SimpleNumberFormatter
    # icu::number::SimpleNumberFormatter::forLocale(
    #       const icu::Locale &locale,
    #       UErrorCode &status
    # )
    fmt1 = SimpleNumberFormatter.for_locale(Locale("de-CH"))
    assert isinstance(fmt1, SimpleNumberFormatter)
    fv = fmt1.format_int64(-1000007)
    assert isinstance(fv, FormattedNumber)
    assert fv.to_temp_string() == "-1’000’007"

    fmt2 = SimpleNumberFormatter.for_locale("de-CH")
    assert isinstance(fmt2, SimpleNumberFormatter)
    fv = fmt2.format_int64(-1000007)
    assert isinstance(fv, FormattedNumber)
    assert fv.to_temp_string() == "-1’000’007"


def test_for_locale_and_grouping_strategy():
    # static SimpleNumberFormatter
    # icu::number::SimpleNumberFormatter::forLocaleAndGroupingStrategy(
    #       const icu::Locale &locale,
    #       UNumberGroupingStrategy groupingStrategy,
    #       UErrorCode &status
    # )
    fmt1 = SimpleNumberFormatter.for_locale_and_grouping_strategy(
        Locale("de-CH"), UNumberGroupingStrategy.UNUM_GROUPING_AUTO
    )
    assert isinstance(fmt1, SimpleNumberFormatter)
    fv = fmt1.format_int64(-1000007)
    assert isinstance(fv, FormattedNumber)
    assert fv.to_temp_string() == "-1’000’007"

    fmt2 = SimpleNumberFormatter.for_locale_and_grouping_strategy(
        "de-CH", UNumberGroupingStrategy.UNUM_GROUPING_AUTO
    )
    assert isinstance(fmt2, SimpleNumberFormatter)
    fv = fmt2.format_int64(-1000007)
    assert isinstance(fv, FormattedNumber)
    assert fv.to_temp_string() == "-1’000’007"


def test_for_locale_and_symbols_and_grouping_strategy():
    # static SimpleNumberFormatter
    # icu::number::SimpleNumberFormatter::forLocaleAndSymbolsAndGroupingStrategy(
    #   	const icu::Locale &locale,
    #       const DecimalFormatSymbols &symbols,
    #       UNumberGroupingStrategy groupingStrategy,
    #       UErrorCode &status
    # )
    symbols = DecimalFormatSymbols("bn")
    fmt1 = SimpleNumberFormatter.for_locale_and_symbols_and_grouping_strategy(
        Locale("en-US"),
        symbols,
        UNumberGroupingStrategy.UNUM_GROUPING_ON_ALIGNED,
    )
    assert isinstance(fmt1, SimpleNumberFormatter)
    fv = fmt1.format_int64(987654321)
    assert isinstance(fv, FormattedNumber)
    assert fv.to_temp_string() == "৯৮৭,৬৫৪,৩২১"

    fmt2 = SimpleNumberFormatter.for_locale_and_symbols_and_grouping_strategy(
        "en-US",
        symbols,
        UNumberGroupingStrategy.UNUM_GROUPING_ON_ALIGNED,
    )
    assert isinstance(fmt2, SimpleNumberFormatter)
    fv = fmt2.format_int64(987654321)
    assert isinstance(fv, FormattedNumber)
    assert fv.to_temp_string() == "৯৮৭,৬৫৪,৩২১"


def test_format_73():
    fmt = SimpleNumberFormatter.for_locale("bn")

    # icu::number::SimpleNumber::SimpleNumber()
    value = SimpleNumber()

    # FormattedNumber icu::number::SimpleNumberFormatter::format(
    #       SimpleNumber value,
    #       UErrorCode &status
    # ) const
    with pytest.raises(ICUError) as exc_info:
        _ = fmt.format(value)  # Default constructed number
    assert exc_info.value.args[0] == UErrorCode.U_ILLEGAL_ARGUMENT_ERROR

    # static SimpleNumber icu::number::SimpleNumber::forInt64(
    #       int64_t value,
    #       UErrorCode &status
    # )
    value = SimpleNumber.for_int64(1250000)
    assert isinstance(value, SimpleNumber)

    # void icu::number::SimpleNumber::multiplyByPowerOfTen(
    #       int32_t power,
    #       UErrorCode &status
    # )
    value.multiply_by_power_of_ten(-2)

    # void icu::number::SimpleNumber::roundTo(
    #       int32_t power,
    #       UNumberFormatRoundingMode roundingMode,
    #       UErrorCode &status
    # )
    value.round_to(3, UNumberFormatRoundingMode.UNUM_ROUND_HALFUP)

    # void icu::number::SimpleNumber::setMinimumFractionDigits(
    #       uint32_t minimumFractionDigits,
    #       UErrorCode &status
    # )
    value.set_minimum_fraction_digits(2)

    # void icu::number::SimpleNumber::setMinimumIntegerDigits(
    #       uint32_t minimumIntegerDigits,
    #       UErrorCode &status
    # )
    value.set_minimum_integer_digits(6)

    # void icu::number::SimpleNumber::setSign(
    #       USimpleNumberSign sign,
    #       UErrorCode &status
    # )
    value.set_sign(USimpleNumberSign.UNUM_SIMPLE_NUMBER_NO_SIGN)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 75, reason="ICU4C<75")
def test_format_75():
    snf = SimpleNumberFormatter.for_locale_and_grouping_strategy(
        "en-US", UNumberGroupingStrategy.UNUM_GROUPING_ON_ALIGNED
    )

    # void icu::number::SimpleNumber::setMaximumIntegerDigits(
    #       uint32_t maximumIntegerDigits,
    #       UErrorCode &status
    # )
    num = SimpleNumber.for_int64(918273645)
    num.multiply_by_power_of_ten(-4)
    num.set_maximum_integer_digits(4)
    result = snf.format(num)
    assert result.to_temp_string() == "1,827.3645"

    with pytest.raises(ICUError) as exc_info:
        _ = snf.format(num)  # Use of moved number
    assert exc_info.value.args[0] == UErrorCode.U_ILLEGAL_ARGUMENT_ERROR



import pytest

from icupy.icu import U_ICU_VERSION_MAJOR_NUM

if U_ICU_VERSION_MAJOR_NUM < 72:
    pytest.skip("ICU4C<72", allow_module_level=True)

# fmt: off
from icupy.icu import (
    DisplayOptions, UDisplayOptionsCapitalization,
    UDisplayOptionsDisplayLength, UDisplayOptionsGrammaticalCase,
    UDisplayOptionsNameStyle, UDisplayOptionsNounClass,
    UDisplayOptionsPluralCategory, UDisplayOptionsSubstituteHandling,
    udispopt_from_grammatical_case_identifier,
    udispopt_from_noun_class_identifier,
    udispopt_from_plural_category_identifier,
    udispopt_get_grammatical_case_identifier,
    udispopt_get_noun_class_identifier,
    udispopt_get_plural_category_identifier,
)

# fmt: on


def test_api():
    # UDisplayOptionsGrammaticalCase
    # udispopt_fromGrammaticalCaseIdentifier(const char *identifier)
    result = udispopt_from_grammatical_case_identifier("")
    assert isinstance(result, UDisplayOptionsGrammaticalCase)
    assert (
        result
        == UDisplayOptionsGrammaticalCase.UDISPOPT_GRAMMATICAL_CASE_UNDEFINED
    )

    with pytest.raises(TypeError):
        _ = udispopt_from_grammatical_case_identifier(None)

    # UDisplayOptionsNounClass
    # udispopt_fromNounClassIdentifier(const char *identifier)
    result = udispopt_from_noun_class_identifier("")
    assert isinstance(result, UDisplayOptionsNounClass)
    assert result == UDisplayOptionsNounClass.UDISPOPT_NOUN_CLASS_UNDEFINED

    with pytest.raises(TypeError):
        _ = udispopt_from_noun_class_identifier(None)

    # UDisplayOptionsPluralCategory
    # udispopt_fromPluralCategoryIdentifier(const char *identifier)
    result = udispopt_from_plural_category_identifier("")
    assert isinstance(result, UDisplayOptionsPluralCategory)
    assert (
        result
        == UDisplayOptionsPluralCategory.UDISPOPT_PLURAL_CATEGORY_UNDEFINED
    )

    with pytest.raises(TypeError):
        _ = udispopt_from_plural_category_identifier(None)

    # const char *
    # udispopt_getGrammaticalCaseIdentifier(
    #       UDisplayOptionsGrammaticalCase grammaticalCase
    # )
    result = udispopt_get_grammatical_case_identifier(
        UDisplayOptionsGrammaticalCase.UDISPOPT_GRAMMATICAL_CASE_UNDEFINED
    )
    assert isinstance(result, str)

    # const char *
    # udispopt_getNounClassIdentifier(UDisplayOptionsNounClass nounClass)
    result = udispopt_get_noun_class_identifier(
        UDisplayOptionsNounClass.UDISPOPT_NOUN_CLASS_UNDEFINED
    )
    assert isinstance(result, str)

    # const char *
    # udispopt_getPluralCategoryIdentifier(UDisplayOptionsPluralCategory pluralCategory)
    result = udispopt_get_plural_category_identifier(
        UDisplayOptionsPluralCategory.UDISPOPT_PLURAL_CATEGORY_UNDEFINED
    )
    assert isinstance(result, str)


def test_builder():
    # static icu::DisplayOptions::Builder icu::DisplayOptions::builder()
    bld = DisplayOptions.builder()
    assert isinstance(bld, DisplayOptions.Builder)

    # DisplayOptions Builder::build()
    ops = bld.build()
    assert isinstance(ops, DisplayOptions)

    # UDisplayOptionsCapitalization icu::DisplayOptions::getCapitalization()
    assert (
        ops.get_capitalization()
        == UDisplayOptionsCapitalization.UDISPOPT_CAPITALIZATION_UNDEFINED
    )

    # UDisplayOptionsDisplayLength icu::DisplayOptions::getDisplayLength()
    assert (
        ops.get_display_length()
        == UDisplayOptionsDisplayLength.UDISPOPT_DISPLAY_LENGTH_UNDEFINED
    )

    # UDisplayOptionsGrammaticalCase icu::DisplayOptions::getGrammaticalCase()
    assert (
        ops.get_grammatical_case()
        == UDisplayOptionsGrammaticalCase.UDISPOPT_GRAMMATICAL_CASE_UNDEFINED
    )

    # UDisplayOptionsNameStyle icu::DisplayOptions::getNameStyle()
    assert (
        ops.get_name_style()
        == UDisplayOptionsNameStyle.UDISPOPT_NAME_STYLE_UNDEFINED
    )

    # UDisplayOptionsNounClass icu::DisplayOptions::getNounClass()
    assert (
        ops.get_noun_class()
        == UDisplayOptionsNounClass.UDISPOPT_NOUN_CLASS_UNDEFINED
    )

    # UDisplayOptionsPluralCategory icu::DisplayOptions::getPluralCategory()
    assert (
        ops.get_plural_category()
        == UDisplayOptionsPluralCategory.UDISPOPT_PLURAL_CATEGORY_UNDEFINED
    )

    # UDisplayOptionsSubstituteHandling icu::DisplayOptions::getSubstituteHandling()
    assert (
        ops.get_substitute_handling()
        == UDisplayOptionsSubstituteHandling.UDISPOPT_SUBSTITUTE_HANDLING_UNDEFINED
    )

    # Builder &Builder::setCapitalization(
    #       UDisplayOptionsCapitalization capitalization
    # )
    result = bld.set_capitalization(
        UDisplayOptionsCapitalization.UDISPOPT_CAPITALIZATION_BEGINNING_OF_SENTENCE
    )
    assert id(result) == id(bld)

    # Builder &Builder::setDisplayLength(
    #       UDisplayOptionsDisplayLength displayLength
    # )
    result = bld.set_display_length(
        UDisplayOptionsDisplayLength.UDISPOPT_DISPLAY_LENGTH_FULL
    )
    assert id(result) == id(bld)

    # Builder &Builder::setGrammaticalCase(
    #       UDisplayOptionsGrammaticalCase grammaticalCase
    # )
    result = bld.set_grammatical_case(
        UDisplayOptionsGrammaticalCase.UDISPOPT_GRAMMATICAL_CASE_ABLATIVE
    )
    assert id(result) == id(bld)

    # Builder &Builder::setNameStyle(
    #       UDisplayOptionsNameStyle nameStyle
    # )
    result = bld.set_name_style(
        UDisplayOptionsNameStyle.UDISPOPT_NAME_STYLE_STANDARD_NAMES
    )
    assert id(result) == id(bld)

    # Builder &Builder::setNounClass(
    #       UDisplayOptionsNounClass nounClass
    # )
    result = bld.set_noun_class(
        UDisplayOptionsNounClass.UDISPOPT_NOUN_CLASS_PERSONAL
    )
    assert id(result) == id(bld)

    # Builder &Builder::setPluralCategory(
    #       UDisplayOptionsPluralCategory pluralCategory
    # )
    result = bld.set_plural_category(
        UDisplayOptionsPluralCategory.UDISPOPT_PLURAL_CATEGORY_FEW
    )
    assert id(result) == id(bld)

    # Builder &Builder::setSubstituteHandling(
    #       UDisplayOptionsSubstituteHandling substituteHandling
    # )
    result = bld.set_substitute_handling(
        UDisplayOptionsSubstituteHandling.UDISPOPT_SUBSTITUTE_HANDLING_NO_SUBSTITUTE
    )
    assert id(result) == id(bld)

    ops2 = bld.build()
    assert id(ops2) != id(ops)
    assert (
        ops2.get_capitalization()
        == UDisplayOptionsCapitalization.UDISPOPT_CAPITALIZATION_BEGINNING_OF_SENTENCE
    )
    assert (
        ops2.get_display_length()
        == UDisplayOptionsDisplayLength.UDISPOPT_DISPLAY_LENGTH_FULL
    )
    assert (
        ops2.get_grammatical_case()
        == UDisplayOptionsGrammaticalCase.UDISPOPT_GRAMMATICAL_CASE_ABLATIVE
    )
    assert (
        ops2.get_name_style()
        == UDisplayOptionsNameStyle.UDISPOPT_NAME_STYLE_STANDARD_NAMES
    )
    assert (
        ops2.get_noun_class()
        == UDisplayOptionsNounClass.UDISPOPT_NOUN_CLASS_PERSONAL
    )
    assert (
        ops2.get_plural_category()
        == UDisplayOptionsPluralCategory.UDISPOPT_PLURAL_CATEGORY_FEW
    )
    assert (
        ops2.get_substitute_handling()
        == UDisplayOptionsSubstituteHandling.UDISPOPT_SUBSTITUTE_HANDLING_NO_SUBSTITUTE
    )

    # Builder icu::DisplayOptions::copyToBuilder()
    bld2 = ops2.copy_to_builder()
    assert isinstance(bld2, DisplayOptions.Builder)

    ops3 = bld2.build()
    assert id(ops3) != id(ops2)
    assert id(ops3) != id(ops)
    assert (
        ops3.get_capitalization()
        == UDisplayOptionsCapitalization.UDISPOPT_CAPITALIZATION_BEGINNING_OF_SENTENCE
    )
    assert (
        ops3.get_display_length()
        == UDisplayOptionsDisplayLength.UDISPOPT_DISPLAY_LENGTH_FULL
    )
    assert (
        ops3.get_grammatical_case()
        == UDisplayOptionsGrammaticalCase.UDISPOPT_GRAMMATICAL_CASE_ABLATIVE
    )
    assert (
        ops3.get_name_style()
        == UDisplayOptionsNameStyle.UDISPOPT_NAME_STYLE_STANDARD_NAMES
    )
    assert (
        ops3.get_noun_class()
        == UDisplayOptionsNounClass.UDISPOPT_NOUN_CLASS_PERSONAL
    )
    assert (
        ops3.get_plural_category()
        == UDisplayOptionsPluralCategory.UDISPOPT_PLURAL_CATEGORY_FEW
    )
    assert (
        ops3.get_substitute_handling()
        == UDisplayOptionsSubstituteHandling.UDISPOPT_SUBSTITUTE_HANDLING_NO_SUBSTITUTE
    )

    # icu::DisplayOptions::DisplayOptions(const DisplayOptions &other)
    ops4 = DisplayOptions(ops2)
    assert (
        ops4.get_capitalization()
        == UDisplayOptionsCapitalization.UDISPOPT_CAPITALIZATION_BEGINNING_OF_SENTENCE
    )
    assert (
        ops4.get_display_length()
        == UDisplayOptionsDisplayLength.UDISPOPT_DISPLAY_LENGTH_FULL
    )
    assert (
        ops4.get_grammatical_case()
        == UDisplayOptionsGrammaticalCase.UDISPOPT_GRAMMATICAL_CASE_ABLATIVE
    )
    assert (
        ops4.get_name_style()
        == UDisplayOptionsNameStyle.UDISPOPT_NAME_STYLE_STANDARD_NAMES
    )
    assert (
        ops4.get_noun_class()
        == UDisplayOptionsNounClass.UDISPOPT_NOUN_CLASS_PERSONAL
    )
    assert (
        ops4.get_plural_category()
        == UDisplayOptionsPluralCategory.UDISPOPT_PLURAL_CATEGORY_FEW
    )
    assert (
        ops4.get_substitute_handling()
        == UDisplayOptionsSubstituteHandling.UDISPOPT_SUBSTITUTE_HANDLING_NO_SUBSTITUTE
    )



import copy

import pytest

# fmt: off
from icupy.icu import (
    INT32_MAX, U_ICU_VERSION_MAJOR_NUM, USET_ADD_CASE_MAPPINGS,
    USET_CASE_INSENSITIVE, USET_IGNORE_SPACE,
    USET_SERIALIZED_STATIC_ARRAY_CAPACITY, ICUError, ParsePosition, UErrorCode,
    UMatchDegree, UnicodeSet, UnicodeString, UProperty, USetSpanCondition,
)

# fmt: on


def test_add():
    test1 = UnicodeSet()
    assert test1.size() == 0

    # [1]
    # UnicodeSet &icu::UnicodeSet::add(const UnicodeString &s)
    #
    # [2]
    # UnicodeSet &icu::UnicodeSet::add(UChar32 c)
    #
    # [3]
    # UnicodeSet &icu::UnicodeSet::add(UChar32 start, UChar32 end)
    result = test1.add(UnicodeString("ab")).add(0xDF).add(0x30, 0x39)
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 12  # [0-9\u00DF{ab}]

    result = test1.add("cd").add("ef")
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 14  # [0-9ab\u00DF{ab}{cd}{ef}]


def test_add_all():
    test1 = UnicodeSet()
    assert test1.size() == 0
    test2 = UnicodeSet(UnicodeString("[\u00df{ab}]"))
    assert test2.size() == 2

    # [1]
    # UnicodeSet &icu::UnicodeSet::addAll(const UnicodeSet &c)
    #
    # [2]
    # UnicodeSet &icu::UnicodeSet::addAll(const UnicodeString &s)
    result = test1.add_all(test2).add_all(UnicodeString("0123456789"))
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 2 + 10  # [0-9\u00DF{ab}]

    result = test1.add_all("cd").add_all("ef")
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 2 + 10 + 4  # [0-9c-f\u00DF{ab}]


def test_add_match_set_to():
    test1 = UnicodeSet(UnicodeString("[0-9]"))
    test2 = UnicodeSet(UnicodeString("[a-z]"))
    assert test1.size() == 10
    assert test2.size() == 26

    # void icu::UnicodeSet::addMatchSetTo(UnicodeSet &toUnionTo)
    test1.add_match_set_to(test2)
    assert test1.size() == 10  # [0-9]
    assert test1.contains(0x30, 0x39)
    assert test2.size() == 10 + 26  # [0-9a-z]
    assert test2.contains(0x30, 0x39)
    assert test2.contains(0x61, 0x7A)


def test_api():
    assert UnicodeSet.IGNORE_SPACE == USET_IGNORE_SPACE
    assert UnicodeSet.CASE_INSENSITIVE == USET_CASE_INSENSITIVE
    assert UnicodeSet.ADD_CASE_MAPPINGS == USET_ADD_CASE_MAPPINGS
    assert (
        UnicodeSet.SERIALIZED_STATIC_ARRAY_CAPACITY
        == USET_SERIALIZED_STATIC_ARRAY_CAPACITY
    )

    test1 = UnicodeSet(0x30, 0x39)

    # int32_t icu::UnicodeSet::getRangeCount(void)
    assert test1.get_range_count() == 1

    # UChar32 icu::UnicodeSet::getRangeEnd(int32_t index)
    assert test1.get_range_end(0) == 0x39

    # UChar32 icu::UnicodeSet::getRangeStart(int32_t index)
    assert test1.get_range_start(0) == 0x30

    # int32_t icu::UnicodeSet::indexOf(UChar32 c)
    assert test1.index_of(0x39) == 9
    assert test1.index_of(0x61) == -1

    # UBool icu::UnicodeSet::isBogus(void)
    assert test1.is_bogus() is False

    # UBool icu::UnicodeSet::isEmpty(void)
    assert test1.is_empty() is False

    # UBool icu::UnicodeSet::isFrozen()
    assert test1.is_frozen() is False

    # int32_t icu::UnicodeSet::size(void)
    assert test1.size() == 10
    assert len(test1) == 10

    # UnicodeSet &icu::UnicodeSet::compact()
    result = test1.compact()
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)

    # UnicodeSet &icu::UnicodeSet::clear(void)
    result = test1.clear()
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)

    assert test1.is_bogus() is False
    assert test1.is_empty() is True
    assert test1.is_frozen() is False
    assert test1.size() == 0
    assert len(test1) == 0

    # UnicodeSet *icu::UnicodeSet::freeze()
    test1.add(0x30, 0x39)
    result = test1.freeze()
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.is_bogus() is False
    assert test1.is_empty() is False
    assert test1.is_frozen() is True
    assert test1.size() == 10
    assert len(test1) == 10

    # static UBool icu::UnicodeSet::resemblesPattern(
    #       const UnicodeString &pattern,
    #       int32_t pos
    # )
    pattern = UnicodeString("[0-9\u00df{ab}]")
    assert UnicodeSet.resembles_pattern(pattern, 0) is True
    assert UnicodeSet.resembles_pattern(pattern, 1) is False

    assert UnicodeSet.resembles_pattern("[0-9\u00df{ab}]", 0) is True
    assert UnicodeSet.resembles_pattern("[0-9\u00df{ab}]", 1) is False

    test2 = UnicodeSet(UnicodeString("[0-9\u00df{ab}]"))
    assert test2.is_bogus() is False
    assert test2.size() == 12

    # UnicodeString &icu::UnicodeSet::toPattern(
    #       UnicodeString &result,
    #       UBool escapeUnprintable = false
    # )
    result = UnicodeString()
    output = test2.to_pattern(result)
    assert isinstance(output, UnicodeString)
    assert id(result) == id(output)
    assert result == "[0-9\xdf{ab}]"

    output = test2.to_pattern(result, False)
    assert isinstance(output, UnicodeString)
    assert id(result) == id(output)
    assert result == "[0-9\xdf{ab}]"

    output = test2.to_pattern(result, True)
    assert isinstance(output, UnicodeString)
    assert id(result) == id(output)
    assert result == "[0-9\\u00DF{ab}]"

    # UnicodeSet &icu::UnicodeSet::set(
    #       UChar32 start,
    #       UChar32 end
    # )
    result = test2.set(0x30, 0x39)
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test2)
    assert test2.size() == 10

    # void icu::UnicodeSet::setToBogus()
    test2.set_to_bogus()
    assert test2.is_bogus() is True

    # UnicodeSet.__repr__() -> str
    assert repr(test1) == "<UnicodeSet('[0-9]')>"

    test3 = UnicodeSet(0, 16).add_all("'\"\\")
    assert repr(test3) == "<UnicodeSet('[\\u0000-\\u0010\"\\'\\\\]')>"


def test_apply_int_property_value():
    test1 = UnicodeSet()
    assert test1.size() == 0

    # UnicodeSet &icu::UnicodeSet::applyIntPropertyValue(
    #       UProperty prop,
    #       int32_t value,
    #       UErrorCode &ec
    # )
    result = test1.apply_int_property_value(
        UProperty.UCHAR_CANONICAL_COMBINING_CLASS, 8
    )  # 8 = Kana Voicing
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 2  # [\u3099\u309A]
    assert test1.contains(0x3099)
    assert test1.contains(0x309A)

    with pytest.raises(ICUError) as exc_info:
        test1.apply_int_property_value(UProperty.UCHAR_INT_LIMIT, 0)
    assert exc_info.value.args[0] == UErrorCode.U_ILLEGAL_ARGUMENT_ERROR


def test_apply_pattern():
    # [1]
    # UnicodeSet &icu::UnicodeSet::applyPattern(
    #       const UnicodeString &pattern,
    #       ParsePosition &pos,
    #       uint32_t options,
    #       const SymbolTable *symbols,
    #       UErrorCode &status
    # )
    test1 = UnicodeSet()
    pattern = UnicodeString("[a-z{ab}]")
    pos = ParsePosition()
    options = UnicodeSet.IGNORE_SPACE | UnicodeSet.CASE_INSENSITIVE
    result = test1.apply_pattern(pattern, pos, options, None)
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert pos.get_index() == pattern.length()
    assert test1.size() == 26 * 2 + 3  # [A-Za-z\u017F\u212A{ab}]
    assert test1.contains(0x41, 0x5A)
    assert test1.contains(0x61, 0x7A)

    pos = ParsePosition()
    result = test1.apply_pattern("[0-9]", pos, options, None)
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert pos.get_index() == 5
    assert test1.size() == 10  # [0-9]
    assert test1.contains(0x30, 0x39)

    with pytest.raises(ICUError) as exc_info:
        test1.apply_pattern(UnicodeString(), pos, options, None)
    assert exc_info.value.args[0] == UErrorCode.U_MALFORMED_SET

    # [2]
    # UnicodeSet &icu::UnicodeSet::applyPattern(
    #       const UnicodeString &pattern,
    #       UErrorCode &status
    # )
    test2 = UnicodeSet()
    result = test2.apply_pattern(pattern)
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test2)
    assert test2.size() == 26 + 1  # [a-z{ab}]
    assert test2.contains(0x61, 0x7A)

    result = test2.apply_pattern("[0-9]")
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test2)
    assert test2.size() == 10  # [0-9]
    assert test2.contains(0x30, 0x39)

    with pytest.raises(ICUError) as exc_info:
        test2.apply_pattern(UnicodeString())
    assert exc_info.value.args[0] == UErrorCode.U_MALFORMED_SET


def test_apply_property_alias():
    test1 = UnicodeSet()

    # UnicodeSet &icu::UnicodeSet::applyPropertyAlias(
    #       const UnicodeString &prop,
    #       const UnicodeString &value,
    #       UErrorCode &ec
    # )
    result = test1.apply_property_alias(
        UnicodeString("Canonical_Combining_Class"),
        UnicodeString("Kana_Voicing"),
    )
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 2  # [\u3099\u309A]
    assert test1.contains(0x3099)
    assert test1.contains(0x309A)

    with pytest.raises(ICUError) as exc_info:
        test1.apply_property_alias(UnicodeString("ccc"), UnicodeString("-"))
    assert exc_info.value.args[0] == UErrorCode.U_ILLEGAL_ARGUMENT_ERROR

    result = test1.apply_property_alias("ccc", UnicodeString("10"))
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 1
    assert test1.contains(0x05B0)

    result = test1.apply_property_alias(UnicodeString("ccc"), "11")
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 1
    assert test1.contains(0x05B1)

    result = test1.apply_property_alias("ccc", "12")
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 1
    assert test1.contains(0x05B2)


def test_char_at():
    test1 = UnicodeSet(0x30, 0x39)
    assert test1.size() == 10

    # UChar32 icu::UnicodeSet::charAt(int32_t index)
    assert test1.char_at(9) == 0x39
    assert test1.char_at(10) == -1
    assert test1.char_at(-1) == -1

    # # __getitem__(self, index: int) -> int
    # assert test1[9] == 0x39
    # with pytest.raises(IndexError):  # IndexError: list index out of range
    #     _ = test1[10]
    # assert test1[-1] == 0x39

    # # __getitem__(self, slice: slice) -> List[int]
    # assert isinstance(test1[:], UnicodeSet)
    # assert test1[1:4] == UnicodeSet(0x31, 0x33)
    # assert test1[1:6:2] == UnicodeSet("[135]")
    # with pytest.raises(ValueError):  # ValueError: slice step cannot be zero
    #     _ = test1[::0]


def test_clone():
    test1 = UnicodeSet(0x30, 0x39)
    assert not test1.is_frozen()

    # UnicodeSet *icu::UnicodeSet::clone()
    test2 = test1.clone()
    assert isinstance(test2, UnicodeSet)
    assert test2 == test1
    assert not test2.is_frozen()

    test1.freeze()
    test3 = test1.clone()
    assert test3 == test1
    assert test1.is_frozen()
    assert test3.is_frozen()

    # UnicodeSet *icu::UnicodeSet::cloneAsThawed()
    test4 = test1.clone_as_thawed()
    assert isinstance(test4, UnicodeSet)
    assert test4 == test1
    assert test1.is_frozen()
    assert not test4.is_frozen()

    test5 = copy.copy(test1)
    assert test5 == test1
    assert test5.is_frozen()

    test6 = copy.deepcopy(test1)
    assert test6 == test1
    assert test6.is_frozen()


def test_close_over():
    pattern = UnicodeString("[aq\u00df{Bc}{bC}{Fi}]")
    test1 = UnicodeSet(pattern)
    test2 = test1.clone()

    # UnicodeSet &icu::UnicodeSet::closeOver(int32_t attribute)
    result = test2.close_over(
        UnicodeSet.CASE_INSENSITIVE | UnicodeSet.ADD_CASE_MAPPINGS
    )
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test2)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 76, reason="ICU4C<76")
def test_code_points():
    # from icu/source/test/intltest/usettest.cpp
    #  UnicodeSetTest::TestCodePointIterator()
    us = UnicodeSet("[abcçカ🚴]")

    # U_HEADER_NESTED_NAMESPACE::USetCodePoints
    # icu::UnicodeSet::codePoints() const
    it = us.code_points()
    assert [UnicodeString(x) for x in it] == ["a", "b", "c", "ç", "カ", "🚴"]


def test_complement():
    test1 = UnicodeSet(UnicodeSet.MIN_VALUE, UnicodeSet.MAX_VALUE)

    # [1]
    # UnicodeSet &icu::UnicodeSet::complement(const UnicodeString &s)
    #
    # [2]
    # UnicodeSet &icu::UnicodeSet::complement(UChar32 c)
    #
    # [3]
    # UnicodeSet &icu::UnicodeSet::complement(UChar32 start, UChar32 end)
    #
    # [4]
    # UnicodeSet &icu::UnicodeSet::complement()
    result = (
        test1.complement(UnicodeString("ab"))
        .complement("cd")
        .complement(0xDF)
        .complement(0x30, 0x39)
        .complement()
    )
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 13  # [0-9\u00DF{ab}{cd}]
    assert test1.contains(0x30, 0x39)
    assert test1.contains(0xDF)
    assert test1.contains("ab")
    assert test1.contains("cd")


def test_complement_all():
    test1 = UnicodeSet(UnicodeString("[0-9\u00df{ab}]"))
    assert test1.size() == 12
    test2 = UnicodeSet(0x30, 0x39)
    assert test2.size() == 10

    # [1]
    # UnicodeSet &icu::UnicodeSet::complementAll(const UnicodeSet &c)
    #
    # [2]
    # UnicodeSet &icu::UnicodeSet::complementAll(const UnicodeString &s)
    result = (
        test1.complement_all(test2)
        .complement_all(UnicodeString("ab"))
        .complement_all("cd")
    )
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 12 - 10 + 2 + 2  # [a-d\u00DF{ab}]
    assert test1.contains(0x61, 0x64)
    assert test1.contains(0xDF)
    assert test1.contains("ab")


def test_contains():
    test1 = UnicodeSet(0x30, 0x39)

    # [1]
    # UBool icu::UnicodeSet::contains(const UnicodeString &s)
    assert test1.contains(UnicodeString("0")) is True
    assert test1.contains(UnicodeString("01")) is False

    assert test1.contains("0") is True
    assert test1.contains("01") is False

    assert UnicodeString("0") in test1
    assert UnicodeString("01") not in test1

    assert "0" in test1
    assert "01" not in test1

    # [2]
    # UBool icu::UnicodeSet::contains(UChar32 c)
    assert test1.contains(0x30) is True
    assert test1.contains(0x61) is False

    assert 0x30 in test1
    assert 0x61 not in test1

    # [3]
    # UBool icu::UnicodeSet::contains(UChar32 start, UChar32 end)
    assert test1.contains(0x30, 0x39) is True
    assert test1.contains(0x61, 0x61) is False


def test_contains_all():
    test1 = UnicodeSet(0x30, 0x39)

    # [1]
    # UBool icu::UnicodeSet::containsAll(const UnicodeSet &c)
    assert test1.contains_all(UnicodeSet(0x30, 0x39)) is True
    assert test1.contains_all(UnicodeSet(0x2F, 0x3A)) is False

    # [2]
    # UBool icu::UnicodeSet::containsAll(const UnicodeString &s)
    assert test1.contains_all(UnicodeString("0123456789")) is True
    assert test1.contains_all(UnicodeString("/0123456789:")) is False

    assert test1.contains_all("0123456789") is True
    assert test1.contains_all("/0123456789:") is False


def test_contains_none():
    test1 = UnicodeSet(0x30, 0x39)

    # [1]
    # UBool icu::UnicodeSet::containsNone(const UnicodeSet &c)
    assert test1.contains_none(UnicodeSet(0x2F, 0x3A)) is False
    assert test1.contains_none(UnicodeSet(0x2F, 0x2F)) is True

    # [2]
    # UBool icu::UnicodeSet::containsNone(const UnicodeString &s)
    assert test1.contains_none(UnicodeString("/0123456789:")) is False
    assert test1.contains_none(UnicodeString("/")) is True

    assert test1.contains_none("/0123456789:") is False
    assert test1.contains_none("/") is True

    # [3]
    # UBool icu::UnicodeSet::containsNone(UChar32 start, UChar32 end)
    assert test1.contains_none(0x2F, 0x3A) is False
    assert test1.contains_none(0x2F, 0x2F) is True


def test_contains_some():
    test1 = UnicodeSet(0x30, 0x39)

    # [1]
    # UBool icu::UnicodeSet::containsSome(const UnicodeSet &s)
    assert test1.contains_some(UnicodeSet(UnicodeString("[0-9{ab}]"))) is True
    assert test1.contains_some(UnicodeSet(UnicodeString("[a-z{ab}]"))) is False

    # [2]
    # UBool icu::UnicodeSet::containsSome(const UnicodeString &s)
    assert test1.contains_some(UnicodeString("ab0123456789")) is True
    assert test1.contains_some(UnicodeString("ab")) is False

    assert test1.contains_some("ab0123456789") is True
    assert test1.contains_some("ab") is False

    # [3]
    # UBool icu::UnicodeSet::containsSome(UChar32 start, UChar32 end)
    assert test1.contains_some(0x2F, 0x3A) is True
    assert test1.contains_some(0x2F, 0x2F) is False


def test_create_from():
    # static UnicodeSet *icu::UnicodeSet::createFrom(const UnicodeString &s)
    test1 = UnicodeSet.create_from(UnicodeString("ab"))
    assert isinstance(test1, UnicodeSet)
    assert test1.size() == 1  # [{ab}]
    assert test1.contains(UnicodeString("ab"))

    test1a = UnicodeSet.create_from("ab")
    assert isinstance(test1a, UnicodeSet)
    assert test1a.size() == 1  # [{ab}]
    assert test1a.contains(UnicodeString("ab"))


def test_create_from_all():
    # static UnicodeSet *icu::UnicodeSet::createFromAll(const UnicodeString &s)
    test1 = UnicodeSet.create_from_all(UnicodeString("ab"))
    assert isinstance(test1, UnicodeSet)
    assert test1.size() == 2  # [ab]
    assert test1.contains(0x61, 0x62)

    test1 = UnicodeSet.create_from_all("ab")
    assert isinstance(test1, UnicodeSet)
    assert test1.size() == 2  # [ab]
    assert test1.contains(0x61, 0x62)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 63, reason="ICU4C<63")
def test_from_uset():
    from icupy.icu import u_get_binary_property_set

    # [1]
    # const UnicodeSet *icu::UnicodeSet::fromUSet(const USet *uset)
    uset1 = u_get_binary_property_set(UProperty.UCHAR_POSIX_BLANK)
    test1 = UnicodeSet.from_uset(uset1)
    assert isinstance(test1, UnicodeSet)
    assert test1.size() > 0
    assert test1.contains(0x20)  # U+0020: Space (SP)

    # USet *icu::UnicodeSet::toUSet()
    uset2 = test1.to_uset()

    # [2]
    # UnicodeSet *icu::UnicodeSet::fromUSet(USet *uset)
    test2 = UnicodeSet.from_uset(uset2)
    assert isinstance(test2, UnicodeSet)
    assert test2 == test1


def test_hash_code():
    test1 = UnicodeSet(0x30, 0x39)
    test2 = UnicodeSet()

    # int32_t icu::UnicodeSet::hashCode(void)
    assert test1.hash_code() != test2.hash_code()

    assert hash(test1) == test1.hash_code()
    assert hash(test2) == test2.hash_code()

    test2.add(0x30, 0x39)
    assert test1.hash_code() == test2.hash_code()

    assert hash(test1) == test1.hash_code()
    assert hash(test2) == test2.hash_code()


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 70, reason="ICU4C<70")
def test_has_strings():
    test1 = UnicodeSet("[a-z]")
    test2 = UnicodeSet("[a-z{ab}]")

    # UBool icu::UnicodeSet::hasStrings()
    assert test1.has_strings() is False
    assert test2.has_strings() is True


def test_iter():
    # from icu/source/test/intltest/usettest.cpp
    #  UnicodeSetTest::TestElementIterator()
    us = UnicodeSet("[abcçカ🚴{}{abc}{de}]")

    result = UnicodeString()
    [result.append(' "').append(el).append('"') for el in us]
    if U_ICU_VERSION_MAJOR_NUM < 76:
        assert result == ' "a" "b" "c" "ç" "カ" "🚴" "" "" ""'
    else:
        assert result == ' "a" "b" "c" "ç" "カ" "🚴" "" "abc" "de"'

    us = UnicodeSet("[a-cA-D{}{abc}{de}]")
    assert all(isinstance(x, UnicodeString) for x in us)
    assert us[0] == "A"
    assert us[1] == "B"
    assert us[2] == "C"
    assert us[3] == "D"
    assert us[4] == "a"
    assert us[5] == "b"
    assert us[6] == "c"
    if U_ICU_VERSION_MAJOR_NUM < 76:
        assert len(us[7]) == 0
        assert len(us[8]) == 0
        assert len(us[9]) == 0
    else:
        assert len(us[7]) == 0
        assert us[8] == "abc"
        assert us[9] == "de"

    with pytest.raises(IndexError):
        _ = us[10]  # IndexError: elements index out of range

    if U_ICU_VERSION_MAJOR_NUM < 76:
        assert len(us[-1]) == 0
    else:
        assert us[-1] == "de"


def test_matches():
    test1 = UnicodeSet(UnicodeString("[0-9{abc}]"))
    text = UnicodeString("abcd 789")

    # UMatchDegree icu::UnicodeSet::matches(
    #       const Replaceable &text,
    #       int32_t &offset,
    #       int32_t limit,
    #       UBool incremental
    # )
    offset = 0
    result, offset = test1.matches(text, offset, text.length(), True)
    assert result == UMatchDegree.U_MATCH
    assert offset == 3

    result, offset = test1.matches(text, offset, text.length(), True)
    assert result == UMatchDegree.U_MISMATCH
    assert offset == 3

    offset = 5
    result, offset = test1.matches(text, offset, text.length(), True)
    assert result == UMatchDegree.U_MATCH
    assert offset == 6


def test_operator():
    test1 = UnicodeSet(0x30, 0x39)
    test2 = UnicodeSet(UnicodeString("[0-9]"))
    test3 = UnicodeSet(UnicodeString("[0-9{ab}]"))

    # UBool icu::UnicodeSet::operator!=(const UnicodeSet &o)
    assert not (test1 != test2)
    assert test1 != test3
    assert test2 != test3

    # UBool icu::UnicodeSet::operator==(const UnicodeSet &o)
    assert test1 == test2
    assert not (test1 == test3)
    assert not (test2 == test3)

    # UnicodeSet.__eq__(_USetPtr) -> bool
    assert test1 == test2.to_uset()
    assert not (test1 == test3.to_uset())
    assert not (test2 == test3.to_uset())

    assert not (test1 != test2.to_uset())
    assert test1 != test3.to_uset()
    assert test2 != test3.to_uset()


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 63, reason="ICU4C<63")
def test_operator_63():
    from icupy.icu import u_get_binary_property_set

    # UnicodeSet.__eq__(_ConstUSetPtr) -> bool
    uset = u_get_binary_property_set(UProperty.UCHAR_POSIX_BLANK)
    test1 = UnicodeSet.from_uset(uset)
    assert test1 == uset
    assert not (test1 != uset)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 76, reason="ICU4C<76")
def test_ranges():
    # from icu/source/test/intltest/usettest.cpp
    #  UnicodeSetTest::TestRangeIterator()
    us = UnicodeSet("[abcçカ🚴]")

    # U_HEADER_NESTED_NAMESPACE::USetRanges icu::UnicodeSet::ranges() const
    ranges = [(it.range_start, it.range_end) for it in us.ranges()]
    result = UnicodeString()
    [
        result.append(" ").append(start).append("-").append(end)
        for (start, end) in ranges
    ]
    assert result == " a-c ç-ç カ-カ 🚴-🚴"


def test_remove():
    test1 = UnicodeSet(UnicodeString("[0-9\u00df{ab}]"))
    assert test1.size() == 12

    # [1]
    # UnicodeSet &icu::UnicodeSet::remove(const UnicodeString &s)
    #
    # [2]
    # UnicodeSet &icu::UnicodeSet::remove(UChar32 c)
    #
    # [3]
    # UnicodeSet &icu::UnicodeSet::remove(UChar32 start, UChar32 end)
    result = test1.remove(UnicodeString("ab")).remove(0xDF).remove(0x30, 0x39)
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 0

    test1 = UnicodeSet(UnicodeString("[0-9\u00df{ab}]"))
    result = test1.remove("ab").remove(0xDF).remove(0x30, 0x39)
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 0


def test_remove_all():
    test1 = UnicodeSet(UnicodeString("[0-9\u00df{ab}]"))
    test2 = test1.clone()
    assert test2.size() == 12
    test3 = UnicodeSet(0x30, 0x39)

    # [1]
    # UnicodeSet &icu::UnicodeSet::removeAll(const UnicodeSet &c)
    #
    # [2]
    # UnicodeSet &icu::UnicodeSet::removeAll(const UnicodeString &s)
    result = test2.remove_all(test3).remove_all(UnicodeString("ab"))
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test2)
    assert test2.size() == 12 - 10  # [\u00DF{ab}]
    assert test2.contains(0xDF)
    assert test2.contains(UnicodeString("ab"))

    test2 = test1.clone()
    result = test2.remove_all(test3).remove_all("ab")
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test2)
    assert test2.size() == 12 - 10  # [\u00DF{ab}]
    assert test2.contains(0xDF)
    assert test2.contains(UnicodeString("ab"))


def test_remove_all_strings():
    test1 = UnicodeSet(UnicodeString("[0-9\u00df{ab}]"))
    assert test1.size() == 12

    # UnicodeSet &icu::UnicodeSet::removeAllStrings()
    result = test1.remove_all_strings()
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 11  # [0-9\u00DF]
    assert test1.contains(0x30, 0x39)
    assert test1.contains(0xDF)
    assert not test1.contains("ab")


def test_retain():
    test1 = UnicodeSet(UnicodeString("[0-9\u00df{ab}]"))
    test2 = test1.clone()
    test3 = test1.clone()
    assert test1.size() == 12

    # [1]
    # UnicodeSet &icu::UnicodeSet::retain(UChar32 c)
    result = test2.retain(0xDF)
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test2)
    assert test2.size() == 1 + 1  # [\u00DF{ab}]
    assert test1.contains(0xDF)

    # [2]
    # UnicodeSet &icu::UnicodeSet::retain(UChar32 start, UChar32 end)
    result = test3.retain(0x30, 0x39)
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test3)
    assert test3.size() == 10 + 1  # [0-9{ab}]
    assert test1.contains(0x30, 0x39)
    assert test1.contains(UnicodeString("ab"))


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 69, reason="ICU4C<69")
def test_retain_69():
    test1 = UnicodeSet("[0-9\u00df{ab}]")
    test2 = test1.clone()
    test3 = test1.clone()
    assert test1.size() == 12

    # [1]
    # UnicodeSet &icu::UnicodeSet::retain(const UnicodeString &s)
    result = test2.retain(UnicodeString("\xdf"))
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test2)
    assert test2.size() == 1 + 1  # [\u00DF{ab}]
    assert test2.contains(0xDF)
    assert test2.contains("ab")

    result = test3.retain("ab")
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test3)
    assert test3.size() == 1  # [{ab}]
    assert test3.contains("ab")


def test_retain_all():
    test1 = UnicodeSet(UnicodeString("[0-9\u00df{ab}]"))
    test2 = test1.clone()
    test3 = test1.clone()
    assert test1.size() == 12

    # [1]
    # UnicodeSet &icu::UnicodeSet::retainAll(const UnicodeSet &c)
    result = test2.retain_all(UnicodeSet(0xDF, 0xDF))
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test2)
    assert test2.size() == 1  # [\u00DF]
    assert test2.contains(0xDF)

    # [2]
    # UnicodeSet &icu::UnicodeSet::retainAll(const UnicodeString &s)
    result = test3.retain_all(UnicodeString("ab"))
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test3)
    assert test3.size() == 0

    test3 = test1.clone()
    result = test3.retain_all("ab")
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test3)
    assert test3.size() == 0


def test_serialize():
    # int32_t icu::UnicodeSet::serialize(
    #       uint16_t *dest,
    #       int32_t destCapacity,
    #       UErrorCode &ec
    # )
    test1 = UnicodeSet(UnicodeString("[0-9\u00df{ab}]"))
    dest = test1.serialize()
    assert isinstance(dest, list)
    assert dest == [
        4,  # length=((n+2*m)|(m!=0?0x8000:0)) (n=4, m=0)
        ord("0"),
        ord("9") + 1,  # bmp code points
        0xDF,
        0xDF + 1,
    ]

    test2 = UnicodeSet(UnicodeString("[\U0001f000-\U0001f003]"))
    dest = test2.serialize()
    assert isinstance(dest, list)
    assert dest == [
        (0 + 2 * 2) | 0x8000,  # length=((n+2*m)|(m!=0?0x8000:0)) (n=0, m=2)
        0,  # bmpLength
        1,
        0xF000,  # supplementary code points
        (0x1F003 + 1) >> 16,
        (0x1F003 + 1) & 0xFFFF,
    ]


def test_span():
    test1 = UnicodeSet(UnicodeString("[0-9\u00df{ab}]"))

    # [1]
    # int32_t icu::UnicodeSet::span(
    #       const char16_t *s,
    #       int32_t length,
    #       USetSpanCondition spanCondition
    # )
    assert test1.span("abcde", -1, USetSpanCondition.USET_SPAN_CONTAINED) == 2

    # [2]
    # int32_t icu::UnicodeSet::span(
    #       const UnicodeString &s,
    #       int32_t start,
    #       USetSpanCondition spanCondition
    # )
    assert (
        test1.span(
            UnicodeString("abcde"), 0, USetSpanCondition.USET_SPAN_CONTAINED
        )
        == 2
    )

    # int32_t icu::UnicodeSet::spanUTF8(
    #       const char *s,
    #       int32_t length,
    #       USetSpanCondition spanCondition
    # ) const
    b = b"abcde"
    assert test1.span_utf8(b, -1, USetSpanCondition.USET_SPAN_CONTAINED) == 2
    assert (
        test1.span_utf8(b, len(b), USetSpanCondition.USET_SPAN_CONTAINED) == 2
    )
    assert test1.span_utf8(b, 2, USetSpanCondition.USET_SPAN_CONTAINED) == 2
    assert test1.span_utf8(b, 1, USetSpanCondition.USET_SPAN_CONTAINED) == 0


def test_span_back():
    test1 = UnicodeSet(UnicodeString("[0-9\u00df{ab}]"))

    # [1]
    # int32_t icu::UnicodeSet::spanBack(
    #       const char16_t *s,
    #       int32_t length,
    #       USetSpanCondition spanCondition
    # )
    assert (
        test1.span_back("aaaab", -1, USetSpanCondition.USET_SPAN_CONTAINED)
        == 3
    )

    # [2]
    # int32_t icu::UnicodeSet::spanBack(
    #       const UnicodeString &s,
    #       int32_t limit,
    #       USetSpanCondition spanCondition
    # )
    assert (
        test1.span_back(
            UnicodeString("aaaab"),
            INT32_MAX,
            USetSpanCondition.USET_SPAN_CONTAINED,
        )
        == 3
    )


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 76, reason="ICU4C<76")
def test_strings():
    # from icu/source/test/intltest/usettest.cpp
    #  UnicodeSetTest::TestStringIterator()
    us = UnicodeSet("[abcçカ🚴{}{abc}{de}]")

    # U_HEADER_NESTED_NAMESPACE::USetStrings icu::UnicodeSet::strings() const
    it = us.strings()
    result = UnicodeString()
    [result.append(' "').append(s).append('"') for s in it]
    assert result == ' "" "abc" "de"'


def test_unicode_set():
    # [1]
    # icu::UnicodeSet::UnicodeSet()
    test1 = UnicodeSet()
    assert test1.size() == 0

    # [2]
    # icu::UnicodeSet::UnicodeSet(UChar32 start, UChar32 end)
    test2 = UnicodeSet(0x30, 0x39)
    assert test2.size() == 10
    assert test2.contains(0x30, 0x39)

    # [4]
    # icu::UnicodeSet::UnicodeSet(
    #       const UnicodeString &pattern,
    #       UErrorCode &status
    # )
    test4 = UnicodeSet(UnicodeString("[0-9]"))
    assert test4.size() == 10
    assert test4.contains(0x30, 0x39)

    test4a = UnicodeSet("[0-9]")
    assert test4a.size() == 10
    assert test4a.contains(0x30, 0x39)

    # [6]
    # icu::UnicodeSet::UnicodeSet(
    #       const UnicodeString &pattern,
    #       ParsePosition &pos,
    #       uint32_t options,
    #       const SymbolTable *symbols,
    #       UErrorCode &status
    # )
    pattern = UnicodeString("foo [ 0-9 ]")
    pos = ParsePosition()
    pos.set_index(4)
    options = USET_IGNORE_SPACE
    test6 = UnicodeSet(pattern, pos, options, None)
    assert pos.get_index() == pattern.length()
    assert test6.size() == 10
    assert test6.contains(0x30, 0x39)

    pos = ParsePosition()
    pos.set_index(4)
    test6a = UnicodeSet("foo [ 0-9 ]", pos, options, None)
    assert pos.get_index() == 11
    assert test6a.size() == 10
    assert test6a.contains(0x30, 0x39)

    # [7]
    # icu::UnicodeSet::UnicodeSet(const UnicodeSet &o)
    test7 = UnicodeSet(test2)
    assert test7.size() == 10
    assert test7.contains(0x30, 0x39)



import copy
from functools import cmp_to_key

import pytest

# fmt: off
from icupy.icu import (
    U_ICU_VERSION_MAJOR_NUM, CollationElementIterator, CollationKey, Collator,
    Locale, RuleBasedCollator, StringCharacterIterator, StringEnumeration,
    UColAttribute, UColAttributeValue, UColBoundMode, UCollationResult,
    UColReorderCode, UColRuleOption, UnicodeSet, UnicodeString, UScriptCode,
)

# fmt: on


def test_clone():
    test1 = Collator.create_instance(Locale.get_japanese())
    rules1 = test1.get_rules()

    # RuleBasedCollator *icu::RuleBasedCollator::clone()
    test2 = test1.clone()
    assert isinstance(test2, RuleBasedCollator)
    rules2 = test2.get_rules()
    assert rules2 == rules1

    test3 = copy.copy(test1)
    rules3 = test3.get_rules()
    assert rules3 == rules1

    test4 = copy.deepcopy(test1)
    rules4 = test4.get_rules()
    assert rules4 == rules1


def test_compare():
    coll = Collator.create_instance(Locale.get_english())
    source = UnicodeString("ABC")
    target = UnicodeString("abc")

    # [2]
    # UCollationResult icu::Collator::compare(
    #       const char16_t *source,
    #       int32_t sourceLength,
    #       const char16_t *target,
    #       int32_t targetLength,
    #       UErrorCode &status
    # )
    assert coll.compare("ABC", -1, "abc", -1) == UCollationResult.UCOL_GREATER

    # [5]
    # UCollationResult icu::Collator::compare(
    #       const UnicodeString &source,
    #       const UnicodeString &target,
    #       int32_t length,
    #       UErrorCode &status
    # )
    assert coll.compare(source, target, 3) == UCollationResult.UCOL_GREATER
    assert coll.compare("ABC", target, 3) == UCollationResult.UCOL_GREATER
    assert coll.compare(source, "abc", 3) == UCollationResult.UCOL_GREATER
    assert coll.compare("ABC", "abc", 3) == UCollationResult.UCOL_GREATER

    # [6]
    # UCollationResult icu::Collator::compare(
    #       const UnicodeString &source,
    #       const UnicodeString &target,
    #       UErrorCode &status
    # )
    assert coll.compare(source, target) == UCollationResult.UCOL_GREATER
    assert coll.compare("ABC", target) == UCollationResult.UCOL_GREATER
    assert coll.compare(source, "abc") == UCollationResult.UCOL_GREATER
    assert coll.compare("ABC", "abc") == UCollationResult.UCOL_GREATER

    coll.set_attribute(
        UColAttribute.UCOL_STRENGTH, UColAttributeValue.UCOL_PRIMARY
    )
    assert coll.compare("ABC", -1, "abc", -1) == UCollationResult.UCOL_EQUAL
    assert coll.compare(source, target, 3) == UCollationResult.UCOL_EQUAL
    assert coll.compare(source, target) == UCollationResult.UCOL_EQUAL

    # UCollationResult icu::Collator::compareUTF8(
    #       const StringPiece &source,
    #       const StringPiece &target,
    #       UErrorCode &status
    # ) const
    source = "a\ufffdz".encode()
    target = b"a\x80z"
    assert coll.compare_utf8(source, target) == UCollationResult.UCOL_EQUAL


def test_create_collation_element_iterator():
    coll = Collator.create_instance(Locale("es"))

    # [1]
    # CollationElementIterator *
    # icu::RuleBasedCollator::createCollationElementIterator(
    #       const CharacterIterator &source
    # )
    source = StringCharacterIterator(UnicodeString("cha"))
    it1 = coll.create_collation_element_iterator(source)
    assert isinstance(it1, CollationElementIterator)

    order1 = it1.next()
    assert order1 != CollationElementIterator.NULLORDER
    order2 = it1.next()
    assert order2 != CollationElementIterator.NULLORDER
    order3 = it1.next()
    assert order3 != CollationElementIterator.NULLORDER
    assert it1.next() == CollationElementIterator.NULLORDER

    it1.set_offset(3)
    assert it1.get_offset() == 3
    assert it1.previous() == order3
    assert it1.previous() == order2
    assert it1.previous() == order1
    assert it1.previous() == CollationElementIterator.NULLORDER

    assert list(it1) == [order1, order2, order3]

    # [2]
    # CollationElementIterator *
    # icu::RuleBasedCollator::createCollationElementIterator(
    #       const UnicodeString &source
    # )
    source = UnicodeString("cha")
    it2 = coll.create_collation_element_iterator(source)
    assert isinstance(it2, CollationElementIterator)

    order1 = it2.next()
    assert order1 != CollationElementIterator.NULLORDER
    order2 = it2.next()
    assert order2 != CollationElementIterator.NULLORDER
    order3 = it2.next()
    assert order3 != CollationElementIterator.NULLORDER
    assert it2.next() == CollationElementIterator.NULLORDER

    it2.set_offset(3)
    assert it2.get_offset() == 3
    assert it2.previous() == order3
    assert it2.previous() == order2
    assert it2.previous() == order1
    assert it2.previous() == CollationElementIterator.NULLORDER

    assert list(it2) == [order1, order2, order3]

    it2a = coll.create_collation_element_iterator("cha")
    assert isinstance(it2a, CollationElementIterator)

    assert it2a.next() == order1
    assert it2a.next() == order2
    assert it2a.next() == order3
    assert it2a.next() == CollationElementIterator.NULLORDER


def test_create_instance():
    # [1]
    # static Collator *icu::Collator::createInstance(
    #       const Locale &loc,
    #       UErrorCode &err
    # )
    test1 = Collator.create_instance(Locale())
    assert isinstance(test1, RuleBasedCollator)
    rules1 = test1.get_rules()

    test1a = Collator.create_instance(str(Locale.get_default()))
    assert isinstance(test1a, RuleBasedCollator)
    rules1a = test1a.get_rules()
    assert rules1 == rules1a

    # [2]
    # static Collator *icu::Collator::createInstance(UErrorCode &err)
    test2 = Collator.create_instance()
    assert isinstance(test2, RuleBasedCollator)
    rules2 = test2.get_rules()
    assert rules1 == rules2


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 76, reason="ICU4C<76")
def test_equal_to():
    # from icu/source/test/intltest/collationtest.cpp
    #  CollationTest::TestCollatorPredicates()
    coll = Collator.create_instance(Locale.get_root())

    # auto icu::Collator::equal_to() const
    assert coll.equal_to("aaa", "aaa") is True
    assert coll.equal_to(UnicodeString("aaa"), "aaa") is True
    assert coll.equal_to("aaa", UnicodeString("aaa")) is True
    assert coll.equal_to(UnicodeString("aaa"), UnicodeString("aaa")) is True

    assert coll.equal_to("aaa", "bbb") is False


def test_equals():
    coll = Collator.create_instance(Locale.get_english())
    source = UnicodeString("ABC")
    target = UnicodeString("abc")

    # UBool icu::Collator::equals(
    #       const UnicodeString &source,
    #       const UnicodeString &target
    # )
    assert coll.equals(source, target) is False
    assert coll.equals("ABC", target) is False
    assert coll.equals(source, "abc") is False
    assert coll.equals("ABC", "abc") is False

    coll.set_attribute(
        UColAttribute.UCOL_STRENGTH, UColAttributeValue.UCOL_PRIMARY
    )
    assert coll.equals(source, target) is True
    assert coll.equals("ABC", target) is True
    assert coll.equals(source, "abc") is True
    assert coll.equals("ABC", "abc") is True


def test_get_attribute():
    coll = Collator.create_instance(Locale.get_canada_french())

    # UColAttributeValue icu::RuleBasedCollator::getAttribute(
    #       UColAttribute attr,
    #       UErrorCode &status
    # )
    assert (
        coll.get_attribute(UColAttribute.UCOL_FRENCH_COLLATION)
        == UColAttributeValue.UCOL_ON
    )
    assert (
        coll.get_attribute(UColAttribute.UCOL_ALTERNATE_HANDLING)
        == UColAttributeValue.UCOL_NON_IGNORABLE
    )
    assert (
        coll.get_attribute(UColAttribute.UCOL_CASE_FIRST)
        == UColAttributeValue.UCOL_STRENGTH_LIMIT
    )
    assert (
        coll.get_attribute(UColAttribute.UCOL_CASE_LEVEL)
        == UColAttributeValue.UCOL_STRENGTH_LIMIT
    )
    assert (
        coll.get_attribute(UColAttribute.UCOL_NORMALIZATION_MODE)
        == UColAttributeValue.UCOL_STRENGTH_LIMIT
    )
    assert (
        coll.get_attribute(UColAttribute.UCOL_DECOMPOSITION_MODE)
        == UColAttributeValue.UCOL_STRENGTH_LIMIT
    )
    assert (
        coll.get_attribute(UColAttribute.UCOL_STRENGTH)
        == UColAttributeValue.UCOL_TERTIARY
    )
    assert (
        coll.get_attribute(UColAttribute.UCOL_NUMERIC_COLLATION)
        == UColAttributeValue.UCOL_STRENGTH_LIMIT
    )

    # void icu::RuleBasedCollator::setAttribute(
    #       UColAttribute attr,
    #       UColAttributeValue value,
    #       UErrorCode &status
    # )
    coll.set_attribute(
        UColAttribute.UCOL_STRENGTH, UColAttributeValue.UCOL_PRIMARY
    )
    assert (
        coll.get_attribute(UColAttribute.UCOL_STRENGTH)
        == UColAttributeValue.UCOL_PRIMARY
    )


def test_get_available_locales():
    # [1]
    # static const Locale *icu::Collator::getAvailableLocales(
    #       int32_t &count
    # )
    it = Collator.get_available_locales()
    assert isinstance(it, list)
    assert len(it) > 0
    assert all(isinstance(x, Locale) for x in it)
    assert Locale("ja") in it


def test_get_bound():
    coll = Collator.create_instance(Locale("sh"))
    source = coll.get_sort_key("Smith", -1)

    # static int32_t icu::Collator::getBound(
    #       const uint8_t *source,
    #       int32_t sourceLength,
    #       UColBoundMode boundType,
    #       uint32_t noOfLevels,
    #       uint8_t *result,
    #       int32_t resultLength,
    #       UErrorCode &status
    # )
    result1 = Collator.get_bound(
        source, len(source), UColBoundMode.UCOL_BOUND_LOWER, 1
    )
    result2 = Collator.get_bound(
        source, len(source), UColBoundMode.UCOL_BOUND_UPPER, 1
    )
    assert isinstance(result1, bytes)
    assert isinstance(result2, bytes)
    assert len(result1) > 0
    assert len(result2) > 0
    assert result1 < source
    assert result2 > source

    result1a = Collator.get_bound(
        source, -1, UColBoundMode.UCOL_BOUND_LOWER, 1
    )
    result2a = Collator.get_bound(
        source, -1, UColBoundMode.UCOL_BOUND_UPPER, 1
    )
    assert result1 == result1a
    assert result2 == result2a


def test_get_collation_key():
    coll = Collator.create_instance(Locale.get_english())

    # [1]
    # CollationKey &icu::Collator::getCollationKey(
    #       const char16_t *source,
    #       int32_t sourceLength,
    #       CollationKey &key,
    #       UErrorCode &status
    # )
    key1 = CollationKey()
    result = coll.get_collation_key("ABC", -1, key1)
    assert isinstance(result, CollationKey)
    assert id(result) == id(key1)

    # [2]
    # CollationKey &icu::Collator::getCollationKey(
    #       const UnicodeString &source,
    #       CollationKey &key,
    #       UErrorCode &status
    # )
    key2 = CollationKey()
    result = coll.get_collation_key(UnicodeString("abc"), key2)
    assert isinstance(result, CollationKey)
    assert id(result) == id(key2)

    result = coll.get_collation_key("abc", key2)
    assert isinstance(result, CollationKey)
    assert id(result) == id(key2)

    assert key1.compare_to(key2) == UCollationResult.UCOL_GREATER
    assert key2.compare_to(key1) == UCollationResult.UCOL_LESS


def test_get_display_name():
    default_locale = None
    try:
        default_locale = Locale.get_default()
        display_locale = Locale.get_us()
        Locale.set_default(display_locale)
        object_locale = Locale("ja")

        # [1]
        # static UnicodeString &icu::Collator::getDisplayName(
        #       const Locale &objectLocale,
        #       const Locale &displayLocale,
        #       UnicodeString &name
        # )
        name1 = UnicodeString()
        result = Collator.get_display_name(
            object_locale, display_locale, name1
        )
        assert isinstance(result, UnicodeString)
        assert id(result) == id(name1)
        assert name1 == "Japanese"

        name1.remove()
        result = Collator.get_display_name("ja", display_locale, name1)
        assert isinstance(result, UnicodeString)
        assert id(result) == id(name1)
        assert name1 == "Japanese"

        name1.remove()
        result = Collator.get_display_name(object_locale, "en_US", name1)
        assert isinstance(result, UnicodeString)
        assert id(result) == id(name1)
        assert name1 == "Japanese"

        name1.remove()
        result = Collator.get_display_name("ja", "en_US", name1)
        assert isinstance(result, UnicodeString)
        assert id(result) == id(name1)
        assert name1 == "Japanese"

        # [2]
        # static UnicodeString &icu::Collator::getDisplayName(
        #       const Locale &objectLocale,
        #       UnicodeString &name
        # )
        name2 = UnicodeString()
        result = Collator.get_display_name(object_locale, name2)
        assert isinstance(result, UnicodeString)
        assert id(result) == id(name2)
        assert name2 == "Japanese"

        name2.remove()
        result = Collator.get_display_name("ja", name2)
        assert isinstance(result, UnicodeString)
        assert id(result) == id(name2)
        assert name2 == "Japanese"
    finally:
        if default_locale:
            Locale.set_default(default_locale)


def test_get_equivalent_reorder_codes():
    # static int32_t icu::Collator::getEquivalentReorderCodes(
    #       int32_t reorderCode,
    #       int32_t *dest,
    #       int32_t destCapacity,
    #       UErrorCode &status
    # )
    dest = Collator.get_equivalent_reorder_codes(UScriptCode.USCRIPT_KATAKANA)
    assert isinstance(dest, list)
    assert dest == [
        UScriptCode.USCRIPT_HIRAGANA,
        UScriptCode.USCRIPT_KATAKANA,
        UScriptCode.USCRIPT_KATAKANA_OR_HIRAGANA,
    ]


def test_get_functional_equivalent():
    # static Locale icu::Collator::getFunctionalEquivalent(
    #       const char *keyword,
    #       const Locale &locale,
    #       UBool &isAvailable,
    #       UErrorCode &status
    # )
    result, is_available = Collator.get_functional_equivalent(
        "collation", Locale.get_japanese()
    )
    assert isinstance(result, Locale)
    assert result == Locale.get_japanese()
    assert is_available is True

    result, is_available = Collator.get_functional_equivalent(
        "collation", "ja"
    )
    assert isinstance(result, Locale)
    assert result == Locale.get_japanese()
    assert is_available is True


def test_get_keywords():
    # static StringEnumeration *icu::Collator::getKeywords(UErrorCode &status)
    keywords = Collator.get_keywords()
    assert isinstance(keywords, StringEnumeration)
    assert len(keywords) > 0
    assert "collation" in keywords

    # static StringEnumeration *icu::Collator::getKeywordValues(
    #       const char *keyword,
    #       UErrorCode &status
    # )
    values = Collator.get_keyword_values("collation")
    assert isinstance(values, StringEnumeration)
    assert len(values) > 0
    assert "standard" in values

    # static StringEnumeration *icu::Collator::getKeywordValuesForLocale(
    #       const char *keyword,
    #       const Locale &locale,
    #       UBool commonlyUsed,
    #       UErrorCode &status
    # )
    it = Collator.get_keyword_values_for_locale(
        "collation", Locale.get_japanese(), True
    )
    assert isinstance(it, StringEnumeration)
    assert list(it) == ["standard", "unihan", "emoji", "eor", "search"]

    it = Collator.get_keyword_values_for_locale("collation", "ja-JP", True)
    assert isinstance(it, StringEnumeration)
    assert list(it) == ["standard", "unihan", "emoji", "eor", "search"]


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 53, reason="ICU4C<53")
def test_get_max_variable():
    coll = Collator.create_instance(Locale.get_japanese())

    # UColReorderCode icu::Collator::getMaxVariable()
    assert (
        coll.get_max_variable()
        == UColReorderCode.UCOL_REORDER_CODE_PUNCTUATION
    )

    # uint32_t icu::Collator::getVariableTop(UErrorCode &status)
    weight1 = coll.get_variable_top()
    assert isinstance(weight1, int)

    # Collator& icu::Collator::setMaxVariable(
    #       UColReorderCode group,
    #       UErrorCode &errorCode
    # )
    coll.set_max_variable(UColReorderCode.UCOL_REORDER_CODE_CURRENCY)
    assert (
        coll.get_max_variable() == UColReorderCode.UCOL_REORDER_CODE_CURRENCY
    )
    weight2 = coll.get_variable_top()
    assert isinstance(weight2, int)

    assert weight1 != weight2


def test_get_reorder_codes():
    coll = Collator.create_instance(Locale.get_japanese())

    # int32_t icu::Collator::getReorderCodes(
    #       int32_t *dest,
    #       int32_t destCapacity,
    #       UErrorCode &status
    # )
    dest = coll.get_reorder_codes()
    assert isinstance(dest, list)
    assert dest == [
        UScriptCode.USCRIPT_LATIN,
        UScriptCode.USCRIPT_KATAKANA,
        UScriptCode.USCRIPT_HAN,
    ]

    # void icu::Collator::setReorderCodes(
    #       const int32_t *reorderCodes,
    #       int32_t reorderCodesLength,
    #       UErrorCode &status
    # )
    reorder_codes = [
        UScriptCode.USCRIPT_HAN,
        UScriptCode.USCRIPT_KATAKANA,
        UScriptCode.USCRIPT_LATIN,
    ]
    coll.set_reorder_codes(reorder_codes, len(reorder_codes))
    dest = coll.get_reorder_codes()
    assert dest == reorder_codes


def test_get_rules():
    coll = Collator.create_instance(Locale.get_japanese())

    # [1]
    # UnicodeString &icu::RuleBasedCollator::getRules()
    rules1 = coll.get_rules()
    assert isinstance(rules1, UnicodeString)

    # [2]
    # void icu::RuleBasedCollator::getRules(
    #       UColRuleOption delta,
    #       UnicodeString &buffer
    # )
    rules2 = UnicodeString()
    coll.get_rules(UColRuleOption.UCOL_TAILORING_ONLY, rules2)
    assert rules2 == rules1

    rules3 = UnicodeString()
    coll.get_rules(UColRuleOption.UCOL_FULL_RULES, rules3)
    assert rules3 != rules1


def test_get_sort_key():
    coll = Collator.create_instance(Locale.get_japanese())

    # [1]
    # int32_t icu::Collator::getSortKey(
    #       const char16_t *source,
    #       int32_t sourceLength,
    #       uint8_t *result,
    #       int32_t resultLength
    # )
    result1 = coll.get_sort_key("ABC", -1)
    assert isinstance(result1, bytes)
    assert len(result1) > 0

    # [2]
    # int32_t icu::Collator::getSortKey(
    #       const UnicodeString &source,
    #       uint8_t *result,
    #       int32_t resultLength
    # )
    result2 = coll.get_sort_key(UnicodeString("abc"))
    assert isinstance(result2, bytes)
    assert len(result2) > 0

    result2a = coll.get_sort_key("abc")
    assert result2a == result2

    assert result1 > result2


def test_get_tailored_set():
    coll1 = Collator.create_instance(Locale.get_english())

    # UnicodeSet *icu::Collator::getTailoredSet(UErrorCode &status)
    uniset1 = coll1.get_tailored_set()
    assert isinstance(uniset1, UnicodeSet)
    assert len(uniset1) == 0

    coll2 = Collator.create_instance(Locale.get_japanese())
    uniset2 = coll2.get_tailored_set()
    assert isinstance(uniset2, UnicodeSet)
    assert len(uniset2) > 0


def test_get_version():
    coll = Collator.create_instance(Locale.get_japanese())

    # void icu::Collator::getVersion(UVersionInfo info)
    info = coll.get_version()
    assert isinstance(info, tuple)
    assert len(info) == 4
    assert all(isinstance(x, int) for x in info)
    assert info[0] > 0


def test_greater():
    coll = Collator.create_instance(Locale.get_english())
    source = UnicodeString("ABC")
    target = UnicodeString("abc")

    # UBool icu::Collator::greater(
    #       const UnicodeString &source,
    #       const UnicodeString &target
    # )
    assert coll.greater(source, target)
    assert coll.greater("ABC", target)
    assert coll.greater(source, "abc")
    assert coll.greater("ABC", "abc")

    coll.set_attribute(
        UColAttribute.UCOL_STRENGTH, UColAttributeValue.UCOL_PRIMARY
    )
    assert not coll.greater(source, target)
    assert not coll.greater("ABC", target)
    assert not coll.greater(source, "abc")
    assert not coll.greater("ABC", "abc")


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 76, reason="ICU4C<76")
def test_greater_equal():
    # from icu/source/test/intltest/collationtest.cpp
    #  CollationTest::TestCollatorPredicates()
    coll = Collator.create_instance(Locale.get_root())

    # auto icu::Collator::greater_equal() const
    assert coll.greater_equal("aaa", "aaa") is True
    assert coll.greater_equal(UnicodeString("aaa"), "aaa") is True
    assert coll.greater_equal("bbb", UnicodeString("aaa")) is True
    assert (
        coll.greater_equal(UnicodeString("bbb"), UnicodeString("aaa")) is True
    )

    assert coll.greater_equal("aaa", "bbb") is False


def test_greater_or_equal():
    coll = Collator.create_instance(Locale.get_english())
    source = UnicodeString("ABC")
    target = UnicodeString("abc")

    # UBool icu::Collator::greaterOrEqual(
    #       const UnicodeString &source,
    #       const UnicodeString &target
    # )
    assert coll.greater_or_equal(source, target)
    assert coll.greater_or_equal("ABC", target)
    assert coll.greater_or_equal(source, "abc")
    assert coll.greater_or_equal("ABC", "abc")

    coll.set_attribute(
        UColAttribute.UCOL_STRENGTH, UColAttributeValue.UCOL_PRIMARY
    )
    assert coll.greater_or_equal(source, target)
    assert coll.greater_or_equal("ABC", target)
    assert coll.greater_or_equal(source, "abc")
    assert coll.greater_or_equal("ABC", "abc")


def test_hash_code():
    coll1 = Collator.create_instance(Locale.get_english())
    coll2 = Collator.create_instance(Locale("da", "DK", ""))
    coll3 = Collator.create_instance(Locale.get_english())

    # int32_t icu::Collator::hashCode(void)
    assert coll1.hash_code() != coll2.hash_code()
    assert coll1.hash_code() == coll3.hash_code()
    assert coll2.hash_code() != coll3.hash_code()

    assert hash(coll1) == coll1.hash_code()
    assert hash(coll2) == coll2.hash_code()
    assert hash(coll3) == coll3.hash_code()


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 76, reason="ICU4C<76")
def test_less():
    # from icu/source/test/intltest/collationtest.cpp
    #  CollationTest::TestCollatorPredicates()
    coll = Collator.create_instance(Locale.get_root())

    # auto icu::Collator::less() const
    assert coll.less("aaa", "bbb") is True
    assert coll.less(UnicodeString("aaa"), "bbb") is True
    assert coll.less("aaa", UnicodeString("bbb")) is True
    assert coll.less(UnicodeString("aaa"), UnicodeString("bbb")) is True

    assert coll.less("aaa", "aaa") is False
    assert coll.less("bbb", "aaa") is False


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 76, reason="ICU4C<76")
def test_less_equal():
    # from icu/source/test/intltest/collationtest.cpp
    #  CollationTest::TestCollatorPredicates()
    coll = Collator.create_instance(Locale.get_root())

    # auto icu::Collator::less_equal() const
    assert coll.less_equal("aaa", "aaa") is True
    assert coll.less_equal(UnicodeString("aaa"), "aaa") is True
    assert coll.less_equal("aaa", UnicodeString("bbb")) is True
    assert coll.less_equal(UnicodeString("aaa"), UnicodeString("bbb")) is True

    assert coll.less_equal("bbb", "aaa") is False


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 76, reason="ICU4C<76")
def test_not_equal_to():
    # from icu/source/test/intltest/collationtest.cpp
    #  CollationTest::TestCollatorPredicates()
    coll = Collator.create_instance(Locale.get_root())

    # auto icu::Collator::not_equal_to() const
    assert coll.not_equal_to("aaa", "bbb") is True
    assert coll.not_equal_to(UnicodeString("aaa"), "bbb") is True
    assert coll.not_equal_to("aaa", UnicodeString("bbb")) is True
    assert (
        coll.not_equal_to(UnicodeString("aaa"), UnicodeString("bbb")) is True
    )

    assert coll.not_equal_to("aaa", "aaa") is False


def test_operator():
    coll1 = Collator.create_instance(Locale.get_english())
    coll2 = Collator.create_instance(Locale("da", "DK", ""))
    coll3 = Collator.create_instance(Locale.get_english())

    # UBool icu::Collator::operator!=(const Collator &other)
    assert coll1 != coll2

    # UBool icu::Collator::operator==(const Collator &other)
    assert coll1 == coll3
    assert coll2 != coll3


def test_rule_based_collator():
    assert issubclass(RuleBasedCollator, Collator)

    base = Collator.create_instance(Locale.get_japanese())
    rules = base.get_rules()
    assert len(rules) > 0

    # [1]
    # icu::RuleBasedCollator::RuleBasedCollator(
    #       const UnicodeString &rules,
    #       UErrorCode &status
    # )
    test1 = RuleBasedCollator(rules)
    assert test1.get_rules() == rules

    test1a = RuleBasedCollator(str(rules))
    assert test1a.get_rules() == rules

    # [2]
    # icu::RuleBasedCollator::RuleBasedCollator(
    #       const UnicodeString &rules,
    #       ECollationStrength collationStrength,
    #       UErrorCode &status
    # )
    test2 = RuleBasedCollator(rules, Collator.ECollationStrength.TERTIARY)
    assert test2.get_rules() == rules

    test2a = RuleBasedCollator(str(rules), Collator.TERTIARY)
    assert test2a.get_rules() == rules

    # [3]
    # icu::RuleBasedCollator::RuleBasedCollator(
    #       const UnicodeString &rules,
    #       UColAttributeValue decompositionMode,
    #       UErrorCode &status
    # )
    test3 = RuleBasedCollator(rules, UColAttributeValue.UCOL_ON)
    assert test3.get_rules() == rules

    test3a = RuleBasedCollator(str(rules), UColAttributeValue.UCOL_ON)
    assert test3a.get_rules() == rules

    # [4]
    # icu::RuleBasedCollator::RuleBasedCollator(
    #       const UnicodeString &rules,
    #       ECollationStrength collationStrength,
    #       UColAttributeValue decompositionMode,
    #       UErrorCode &status
    # )
    test4 = RuleBasedCollator(
        rules, Collator.ECollationStrength.TERTIARY, UColAttributeValue.UCOL_ON
    )
    assert test4.get_rules() == rules

    test4a = RuleBasedCollator(
        str(rules),
        Collator.ECollationStrength.TERTIARY,
        UColAttributeValue.UCOL_ON,
    )
    assert test4a.get_rules() == rules

    # [6]
    # icu::RuleBasedCollator::RuleBasedCollator(const RuleBasedCollator &other)
    test6 = RuleBasedCollator(test1)
    assert test6.get_rules() == rules

    # int32_t icu::RuleBasedCollator::cloneBinary(
    #       uint8_t *buffer,
    #       int32_t capacity,
    #       UErrorCode &status
    # )
    test1.set_attribute(
        UColAttribute.UCOL_STRENGTH, UColAttributeValue.UCOL_PRIMARY
    )
    buffer = test1.clone_binary()
    assert isinstance(buffer, bytes)
    assert len(buffer) > 0

    # [7]
    # icu::RuleBasedCollator::RuleBasedCollator(
    #       const uint8_t *bin,
    #       int32_t length,
    #       const RuleBasedCollator *base,
    #       UErrorCode &status
    # )
    root = Collator.create_instance(Locale.get_root())
    test7 = RuleBasedCollator(buffer, len(buffer), root)
    assert test1 == test7
    assert (
        test7.get_attribute(UColAttribute.UCOL_STRENGTH)
        == UColAttributeValue.UCOL_PRIMARY
    )

    buffer2 = test7.clone_binary()
    assert len(buffer) == len(buffer2)
    assert buffer == buffer2

    test7a = RuleBasedCollator(buffer2, -1, root)
    assert test1 == test7 == test7a
    assert (
        test7a.get_attribute(UColAttribute.UCOL_STRENGTH)
        == UColAttributeValue.UCOL_PRIMARY
    )

    buffer3 = test7a.clone_binary()
    assert len(buffer) == len(buffer3)
    assert buffer == buffer3


def test_sort():
    # From
    #  https://unicode-org.github.io/icu/userguide/collation/customization/#simple-tailoring-examples
    src = [
        "CUKIĆ RADOJICA",
        "ČUKIĆ SLOBODAN",
        "CUKIĆ SVETOZAR",
        "ČUKIĆ ZORAN",
        "CURIĆ MILOŠ",
        "ĆURIĆ MILOŠ",
        "CVRKALJ ÐURO",
    ]
    base = Collator.create_instance(Locale("sr_Latn"))
    rules = base.get_rules() + UnicodeString("& C < č <<< Č < ć <<< Ć")
    coll = RuleBasedCollator(rules)

    def _cmp(_a, _b):
        nonlocal coll
        _result = coll.compare(_a, -1, _b, -1)
        return (
            -1
            if _result == UCollationResult.UCOL_LESS
            else 0 if _result == UCollationResult.UCOL_EQUAL else 1
        )

    result1 = sorted(src)
    result2 = sorted(src, key=cmp_to_key(_cmp))
    result3 = sorted(src, key=lambda s: coll.get_sort_key(s, -1))  # type: ignore
    assert result1 != result2
    assert result1 != result3
    assert result2 == result3
    assert result3 == [
        "CUKIĆ RADOJICA",
        "CUKIĆ SVETOZAR",
        "CURIĆ MILOŠ",
        "CVRKALJ ÐURO",
        "ČUKIĆ SLOBODAN",
        "ČUKIĆ ZORAN",
        "ĆURIĆ MILOŠ",
    ]



from __future__ import annotations

import sys
from typing import Any

from setuptools import setup

sys.path.append("")
from _build import build  # noqa

setup_kwargs: dict[str, Any] = {}
build(setup_kwargs)
setup(**setup_kwargs)



import pytest

# fmt: off
from icupy.icu import (
    U_SENTINEL, ICUError, StringCharacterIterator, UErrorCode, UnicodeString,
    UnicodeStringVector, UTextVector, utext_char32_at, utext_clone,
    utext_close, utext_copy, utext_current32, utext_equals, utext_extract,
    utext_freeze, utext_get_native_index, utext_get_previous_native_index,
    utext_has_meta_data, utext_is_length_expensive, utext_is_writable,
    utext_move_index32, utext_native_length, utext_next32, utext_next32_from,
    utext_open_character_iterator, utext_open_const_unicode_string,
    utext_open_replaceable, utext_open_uchars, utext_open_unicode_string,
    utext_open_utf8, utext_previous32, utext_previous32_from, utext_replace,
    utext_set_native_index,
)
from icupy.utils import gc

# fmt: on


def test_api():
    s = "\x41\U0001f338\x42"
    with gc(utext_open_uchars(None, s, -1), utext_close) as ut:
        assert ut.magic
        assert ut.flags
        assert ut.provider_properties
        assert ut.p_funcs
        assert utext_char32_at(ut, 0) == 0x41
        assert utext_char32_at(ut, 1) == 0x1F338
        assert utext_char32_at(ut, 2) == 0x1F338
        assert utext_char32_at(ut, 3) == 0x42
        assert utext_char32_at(ut, 4) == U_SENTINEL

        # int32_t utext_extract(UText *ut,
        #                       int64_t nativeStart,
        #                       int64_t nativeLimit,
        #                       UChar *dest,
        #                       int32_t destCapacity,
        #                       UErrorCode *status
        # )
        dest = utext_extract(ut, 1, 3)
        assert isinstance(dest, str)
        assert dest == "\U0001f338"

        utext_set_native_index(ut, 0)
        assert utext_current32(ut) == 0x41
        assert utext_get_native_index(ut) == 0
        assert utext_get_previous_native_index(ut) == 0

        assert utext_has_meta_data(ut) is False
        assert utext_is_length_expensive(ut) is False
        assert utext_is_writable(ut) is False

        assert utext_move_index32(ut, 1) is True
        assert utext_native_length(ut) == 4
        assert utext_next32(ut) == 0x1F338
        assert utext_next32(ut) == 0x42
        assert utext_next32(ut) == U_SENTINEL

        assert utext_next32_from(ut, 2) == 0x1F338
        assert utext_next32(ut) == 0x42

        assert utext_previous32(ut) == 0x42
        assert utext_previous32(ut) == 0x1F338
        assert utext_previous32(ut) == 0x41
        assert utext_previous32(ut) == U_SENTINEL

        assert utext_previous32_from(ut, 3) == 0x1F338
        assert utext_previous32(ut) == 0x41

        result = repr(ut)
        assert result.startswith("<_UTextPtr(<UText(magic=0x")
        start = result.find(",")
        assert start != -1
        assert result.startswith(
            ", flags=0x5, providerProperties=0x4, pFuncs=0x", start
        )
        assert result.endswith(")>)>")


def test_clone_uchars():
    s1 = "ABC"
    s2 = "abc"
    ut1 = utext_open_uchars(None, s1, -1)
    assert not utext_is_writable(ut1)

    # Deep clone
    ut2 = utext_clone(None, ut1, True, True)
    assert ut2 is not None
    assert ut2 != ut1
    assert not utext_is_writable(ut2)
    assert utext_extract(ut2, 0, utext_native_length(ut2)) == "ABC"
    assert utext_close(ut2) is None

    ut2 = utext_open_uchars(None, s2, -1)
    ut3 = utext_clone(ut2, ut1, True, True)
    assert ut3 is not None
    assert ut3 == ut2
    assert ut3 != ut1
    assert not utext_is_writable(ut3)
    assert utext_extract(ut3, 0, utext_native_length(ut3)) == "ABC"
    assert utext_close(ut2) is None
    assert utext_close(ut3) == ut2

    ut2 = utext_open_uchars(None, s2, -1)
    utext_clone(ut2, ut1, True, False)
    assert ut2 != ut1
    assert not utext_is_writable(ut2)
    assert utext_extract(ut2, 0, utext_native_length(ut2)) == "ABC"
    assert utext_close(ut2) is None

    # Shallow clone
    ut4 = utext_clone(None, ut1, False, True)
    assert ut4 is not None
    assert ut4 != ut1
    assert not utext_is_writable(ut4)
    assert utext_extract(ut4, 0, utext_native_length(ut4)) == "ABC"
    assert utext_close(ut4) is None

    ut5 = utext_open_uchars(None, s2, -1)
    ut6 = utext_clone(ut5, ut1, False, True)
    assert ut6 is not None
    assert ut6 == ut5
    assert ut6 != ut1
    assert not utext_is_writable(ut6)
    assert utext_extract(ut6, 0, utext_native_length(ut6)) == "ABC"
    assert utext_close(ut5) is None
    assert utext_close(ut6) == ut5

    ut5 = utext_open_uchars(None, s2, -1)
    utext_clone(ut5, ut1, False, False)
    assert ut5 != ut1
    assert not utext_is_writable(ut5)
    assert utext_extract(ut5, 0, utext_native_length(ut5)) == "ABC"
    assert utext_close(ut5) is None

    assert utext_close(ut1) is None


def test_clone_unicode_string():
    s1 = UnicodeString("ABC")
    s2 = UnicodeString("abc")
    ut1 = utext_open_unicode_string(None, s1)
    assert utext_is_writable(ut1)

    # Deep clone
    ut2 = utext_clone(None, ut1, True, True)
    assert ut2 is not None
    assert ut2 != ut1
    assert not utext_is_writable(ut2)
    assert utext_extract(ut2, 0, utext_native_length(ut2)) == "ABC"
    assert utext_close(ut2) is None

    ut2 = utext_open_unicode_string(None, s2)
    ut3 = utext_clone(ut2, ut1, True, True)
    assert ut3 is not None
    assert ut3 == ut2
    assert ut3 != ut1
    assert not utext_is_writable(ut3)
    assert utext_extract(ut3, 0, utext_native_length(ut3)) == "ABC"
    assert utext_close(ut2) is None
    assert utext_close(ut3) == ut2

    ut2 = utext_open_unicode_string(None, s2)
    utext_clone(ut2, ut1, True, False)
    assert ut2 != ut1
    assert utext_is_writable(ut2)
    assert utext_extract(ut2, 0, utext_native_length(ut2)) == "ABC"
    assert utext_close(ut2) is None

    # Shallow clone
    ut4 = utext_clone(None, ut1, False, True)
    assert ut4 is not None
    assert ut4 != ut1
    assert not utext_is_writable(ut4)
    assert utext_extract(ut4, 0, utext_native_length(ut4)) == "ABC"
    assert utext_close(ut4) is None

    ut5 = utext_open_unicode_string(None, s2)
    ut6 = utext_clone(ut5, ut1, False, True)
    assert ut6 is not None
    assert ut6 == ut5
    assert ut6 != ut1
    assert not utext_is_writable(ut6)
    assert utext_extract(ut6, 0, utext_native_length(ut6)) == "ABC"
    assert utext_close(ut5) is None
    assert utext_close(ut6) == ut5

    ut5 = utext_open_unicode_string(None, s2)
    utext_clone(ut5, ut1, False, False)
    assert ut5 != ut1
    assert utext_is_writable(ut5)
    assert utext_extract(ut5, 0, utext_native_length(ut5)) == "ABC"
    assert utext_close(ut5) is None

    assert utext_close(ut1) is None


def test_copy():
    s1 = "ABC"
    with gc(utext_open_uchars(None, s1, -1), utext_close) as ut1:
        assert not utext_is_writable(ut1)

        with pytest.raises(ICUError) as exc_info:
            utext_copy(ut1, 0, 2, 2, False)
        assert exc_info.value.args[0] == UErrorCode.U_NO_WRITE_PERMISSION

    s2 = UnicodeString(s1)
    with gc(utext_open_unicode_string(None, s2), utext_close) as ut2:
        assert utext_is_writable(ut2)

        # Copy: |AB|C -> AB|AB|C
        utext_copy(ut2, 0, 2, 2, False)
        assert utext_extract(ut2, 0, utext_native_length(ut2)) == "ABABC"

        # Move: A|BA|BC -> ABC|BA|
        utext_copy(ut2, 1, 3, 5, True)
        assert utext_extract(ut2, 0, utext_native_length(ut2)) == "ABCBA"


def test_equals():
    s1 = "ABC"
    with gc(utext_open_uchars(None, s1, -1), utext_close) as ut1:
        with gc(utext_clone(None, ut1, False, True), utext_close) as ut2:
            utext_set_native_index(ut1, 0)
            utext_set_native_index(ut2, 0)
            assert utext_equals(ut1, ut2) is True

            assert utext_next32(ut1) != U_SENTINEL
            assert utext_equals(ut1, ut2) is False

            assert utext_next32(ut2) != U_SENTINEL
            assert utext_equals(ut1, ut2) is True


def test_freeze():
    s = UnicodeString("ABC")
    with gc(utext_open_unicode_string(None, s), utext_close) as ut:
        assert utext_is_writable(ut)
        utext_freeze(ut)
        assert not utext_is_writable(ut)


def test_open_character_iterator():
    src1 = UnicodeString("\x41\U0001f338\x42")
    src2 = UnicodeString("\x61\U0001f338\x62")
    it1 = StringCharacterIterator(src1)
    it2 = StringCharacterIterator(src2)
    with gc(utext_open_character_iterator(None, it1), utext_close) as ut1:
        assert not utext_is_writable(ut1)

        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "A\U0001f338B"

        # Reset
        ut2 = utext_open_character_iterator(ut1, it2)
        assert ut2 == ut1

        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "a\U0001f338b"

        dest2 = utext_extract(ut2, 0, utext_native_length(ut2))
        assert dest2 == "a\U0001f338b"

        # Keep-alive test
        utext_open_character_iterator(ut1, it1)
        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "A\U0001f338B"


def test_open_const_unicode_string():
    s1 = UnicodeString("\x41\U0001f338\x42")
    s2 = UnicodeString("\x61\U0001f338\x62")
    with gc(utext_open_const_unicode_string(None, s1), utext_close) as ut1:
        assert not utext_is_writable(ut1)

        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "A\U0001f338B"

        # Reset
        ut2 = utext_open_const_unicode_string(ut1, s2)
        assert ut2 == ut1

        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "a\U0001f338b"

        dest2 = utext_extract(ut2, 0, utext_native_length(ut2))
        assert dest2 == "a\U0001f338b"

        # Keep-alive test
        utext_open_const_unicode_string(ut1, s1)
        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "A\U0001f338B"


def test_open_replaceable():
    s1 = UnicodeString("\x41\U0001f338\x42")
    s2 = UnicodeString("\x61\U0001f338\x62")
    with gc(utext_open_replaceable(None, s1), utext_close) as ut1:
        assert utext_is_writable(ut1)

        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "A\U0001f338B"

        # Reset
        ut2 = utext_open_replaceable(ut1, s2)
        assert ut2 == ut1

        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "a\U0001f338b"

        dest2 = utext_extract(ut2, 0, utext_native_length(ut2))
        assert dest2 == "a\U0001f338b"

        # Keep-alive test
        utext_open_replaceable(ut1, s1)
        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "A\U0001f338B"


def test_open_uchars():
    s1 = "\x41\U0001f338\x42"
    s2 = "\x61\U0001f338\x62"
    with gc(utext_open_uchars(None, s1, -1), utext_close) as ut1:
        assert not utext_is_writable(ut1)

        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "A\U0001f338B"

        # Reset
        ut2 = utext_open_uchars(ut1, s2, -1)
        assert ut2 == ut1

        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "a\U0001f338b"

        dest2 = utext_extract(ut2, 0, utext_native_length(ut2))
        assert dest2 == "a\U0001f338b"

        # Keep-alive test
        utext_open_uchars(ut1, s1, -1)
        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "A\U0001f338B"

        # NULL string
        utext_open_uchars(ut1, None, -1)
        dest1 = utext_extract(ut1, 0, 4)
        assert len(dest1) == 0


def test_open_unicode_string():
    s1 = UnicodeString("\x41\U0001f338\x42")
    s2 = UnicodeString("\x61\U0001f338\x62")
    with gc(utext_open_unicode_string(None, s1), utext_close) as ut1:
        assert utext_is_writable(ut1)

        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "A\U0001f338B"

        # Reset
        ut2 = utext_open_unicode_string(ut1, s2)
        assert ut2 == ut1

        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "a\U0001f338b"

        dest2 = utext_extract(ut2, 0, utext_native_length(ut2))
        assert dest2 == "a\U0001f338b"

        # Keep-alive test
        utext_open_unicode_string(ut1, s1)
        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "A\U0001f338B"


def test_open_utf8():
    b1 = "\x41\U0001f338\x42".encode()
    b2 = "\x61\U0001f338\x62".encode()
    with gc(utext_open_utf8(None, b1, -1), utext_close) as ut1:
        assert not utext_is_writable(ut1)

        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "A\U0001f338B"

        # Reset
        ut2 = utext_open_utf8(ut1, b2, -1)
        assert ut2 == ut1

        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "a\U0001f338b"

        dest2 = utext_extract(ut2, 0, utext_native_length(ut2))
        assert dest2 == "a\U0001f338b"

        # Keep-alive test
        utext_open_utf8(ut1, b1, -1)
        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "A\U0001f338B"

    with gc(utext_open_utf8(None, b2, len(b2)), utext_close) as ut1:
        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "a\U0001f338b"

    with gc(utext_open_utf8(None, b2, len(b2) - 1), utext_close) as ut1:
        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "a\U0001f338"

    with gc(utext_open_utf8(None, b2, len(b2) - 2), utext_close) as ut1:
        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "a\ufffd"

    with gc(utext_open_utf8(None, b2), utext_close) as ut1:
        dest1 = utext_extract(ut1, 0, utext_native_length(ut1))
        assert dest1 == "a\U0001f338b"


def test_replace():
    s1 = "ABC"
    with gc(utext_open_uchars(None, s1, -1), utext_close) as ut1:
        assert not utext_is_writable(ut1)

        with pytest.raises(ICUError) as exc_info:
            utext_replace(ut1, 1, 2, "abc", -1)
        assert exc_info.value.args[0] == UErrorCode.U_NO_WRITE_PERMISSION

    s2 = UnicodeString(s1)
    with gc(utext_open_unicode_string(None, s2), utext_close) as ut2:
        assert utext_is_writable(ut2)

        assert utext_replace(ut2, 1, 2, "abc", -1) == 2
        assert utext_extract(ut2, 0, utext_native_length(ut2)) == "AabcC"


def test_utext_vector1():
    # [1]
    # UTextVector()
    t = UTextVector()
    assert len(t) == 0

    # UTextVector.append(src: UnicodeString)
    src = [UnicodeString("a"), UnicodeString("b")]
    t.append(src[0])
    t.append(src[1])
    assert len(t) == 2
    assert all(utext_is_writable(x) for x in t)
    assert [utext_extract(x, 0, utext_native_length(x)) for x in t] == [
        "a",
        "b",
    ]

    # Keep a reference (appended UText → UnicodeString)
    assert utext_replace(t[0], 0, utext_native_length(t[0]), "foo", -1) == 2
    assert src == ["foo", "b"]

    # Keep a reference (appended UnicodeString → UText)
    src[1].set_to("bar")
    assert [utext_extract(x, 0, utext_native_length(x)) for x in t] == [
        "foo",
        "bar",
    ]

    # UTextVector.extend(iterable: list[UnicodeString])
    src += [UnicodeString("c"), UnicodeString("d")]
    t.extend(src[2:])
    assert len(t) == 4
    assert all(utext_is_writable(x) for x in t)
    assert [utext_extract(x, 0, utext_native_length(x)) for x in t] == [
        "foo",
        "bar",
        "c",
        "d",
    ]

    # Keep a reference (extended UText → UnicodeString)
    assert utext_replace(t[2], 0, utext_native_length(t[2]), "baz", -1) == 2
    assert src == ["foo", "bar", "baz", "d"]

    # Keep a reference (extended UnicodeString → UText)
    src[3].set_to("qux")
    assert [utext_extract(x, 0, utext_native_length(x)) for x in t] == [
        "foo",
        "bar",
        "baz",
        "qux",
    ]

    # UTextVector.insert(index: int, src: UnicodeString)
    src.insert(0, UnicodeString("a"))
    src.insert(-1, UnicodeString("b"))
    t.insert(0, src[0])
    t.insert(-1, src[-2])
    assert len(t) == 6
    assert all(utext_is_writable(x) for x in t)
    assert [utext_extract(x, 0, utext_native_length(x)) for x in t] == [
        "a",
        "foo",
        "bar",
        "baz",
        "b",
        "qux",
    ]

    with pytest.raises(IndexError):
        t.insert(6, src[0])

    with pytest.raises(IndexError):
        t.insert(-7, src[0])

    # Keep a reference (inserted UText → UnicodeString)
    assert utext_replace(t[0], 0, utext_native_length(t[0]), "0", -1) == 0
    assert src == ["0", "foo", "bar", "baz", "b", "qux"]

    # Keep a reference (inserted UnicodeString → UText)
    src[4].set_to("4")
    assert [utext_extract(x, 0, utext_native_length(x)) for x in t] == [
        "0",
        "foo",
        "bar",
        "baz",
        "4",
        "qux",
    ]

    # UTextVector.__getitem__(index: int) -> _UTextPtr
    with pytest.raises(IndexError):
        _ = t[6]

    with pytest.raises(IndexError):
        _ = t[-7]

    # UTextVector.__delitem__(index: int)
    del t[0]
    del src[0]
    del t[-2]
    del src[-2]
    assert len(t) == 4
    assert [utext_extract(x, 0, utext_native_length(x)) for x in t] == [
        "foo",
        "bar",
        "baz",
        "qux",
    ]

    with pytest.raises(IndexError):
        del t[4]

    with pytest.raises(IndexError):
        del t[-5]

    # UTextVector.__delitem__(index: slice)
    del t[1:4:2]
    assert [utext_extract(x, 0, utext_native_length(x)) for x in t] == [
        "foo",
        "baz",
    ]

    with pytest.raises(ValueError):
        del t[::0]

    # UTextVector.clear()
    t.clear()
    assert len(t) == 0

    # UTextVector.__iadd__(iterable: list[UnicodeString])
    t += src
    assert len(t) == 4
    assert all(utext_is_writable(x) for x in t)
    assert [utext_extract(x, 0, utext_native_length(x)) for x in t] == [
        "foo",
        "bar",
        "baz",
        "qux",
    ]

    # Keep a reference (UText → UnicodeString)
    assert utext_replace(t[0], 0, utext_native_length(t[0]), "a", -1) == -2
    assert src == ["a", "bar", "baz", "qux"]

    # Keep a reference (UnicodeString → UText)
    src[1].set_to("b")
    assert [utext_extract(x, 0, utext_native_length(x)) for x in t] == [
        "a",
        "b",
        "baz",
        "qux",
    ]

    # UTextVector.__reversed__() -> Iterator
    assert reversed(t) is None  # UTextVector does not support reverse iterator


def test_utext_vector2():
    # [2]
    # UTextVector(iterable: list[UnicodeString])
    src = [UnicodeString("a"), UnicodeString("b"), UnicodeString("c")]
    t = UTextVector(src)
    assert len(t) == 3
    assert all(utext_is_writable(x) for x in t)
    assert [utext_extract(x, 0, utext_native_length(x)) for x in t] == [
        "a",
        "b",
        "c",
    ]

    # Keep a reference (initialized UText → UnicodeString)
    assert utext_replace(t[0], 0, utext_native_length(t[0]), "foo", -1) == 2
    assert src == ["foo", "b", "c"]

    # Keep a reference (initialized UnicodeString → UText)
    src[1].set_to("bar")
    assert [utext_extract(x, 0, utext_native_length(x)) for x in t] == [
        "foo",
        "bar",
        "c",
    ]


def test_utext_vector2_2():
    # [2]´
    # UTextVector(iterable: list[UnicodeString])
    src = UnicodeStringVector(3)
    src[0].set_to("a")
    src[1].set_to("b")
    src[2].set_to("c")
    t = UTextVector(src)
    assert len(t) == 3
    assert all(utext_is_writable(x) for x in t)
    assert [utext_extract(x, 0, utext_native_length(x)) for x in t] == [
        "a",
        "b",
        "c",
    ]

    # Keep a reference (initialized UText → UnicodeString)
    assert utext_replace(t[0], 0, utext_native_length(t[0]), "foo", -1) == 2
    assert [str(x) for x in src] == ["foo", "b", "c"]

    # Keep a reference (initialized UnicodeString → UText)
    src[1].set_to("bar")
    assert [utext_extract(x, 0, utext_native_length(x)) for x in t] == [
        "foo",
        "bar",
        "c",
    ]
